# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np

import sys
import os
import cPickle
if __name__ == "__main__":
    sys.path.insert(0, os.path.abspath('../..'))
from labrad import units
ValueArray = units.ValueArray
Value = units.Value

class LabradUnitsTests(unittest.TestCase):
    def testParsing(self):
        # prefixes
        # multiplication
        # division
        # powers
        pass

    def testArithmetic(self):
        m = units.Unit('m')
        kg = units.Unit('kg')
        km = units.Unit('km')

        #self.assertEqual(units.Value(5.0, None)*m, 5.0*m)

        # addition
        self.assertEqual(1.0*kg + 0.0, 1.0*kg)
        with self.assertRaises(TypeError): _ = 1.0*kg + 1.0*m
        with self.assertRaises(TypeError): _ = 1.0*kg + 2.0
        self.assertAlmostEqual(1.0*km/m + 5.0, 1005)
        self.assertNotEqual(1.0*kg, None)

    @unittest.expectedFailure # TODO: fix isfinite on ValueArray
    def testValueArray(self):
        # Slicing
        self.assertTrue((ValueArray([1, 2, 3], 'm')[0:2] == ValueArray([1, 2], 'm')).all())
        # Cast to unit
        self.assertTrue((ValueArray([1.2, 4, 5], 'm')['m'] == np.array([1.2, 4, 5])).all())
        # Addition and subtraction of compatible units
        self.assertTrue((ValueArray([3, 4], 'm') + ValueArray([100, 200], 'cm') ==
                         ValueArray([4, 6], 'm')).all())
        self.assertTrue((ValueArray([2, 3, 4], 'm') - ValueArray([100, 200, 300], 'cm') ==
                         ValueArray([1, 1, 1], 'm')).all())
        # Division with units remaining
        self.assertTrue((ValueArray([3, 4, 5], 'm') / ValueArray([1, 2, 5], 's') ==
                         ValueArray([3, 2, 1], 'm/s')).all())
        # Division with no units remaining
        self.assertTrue((ValueArray([3, 4, 5], 'm') / ValueArray([1, 2, 5], 'm') ==
                         ValueArray([3, 2, 1], '')).all())
        # Powers
        self.assertTrue((ValueArray([2, 3], 'm')**2 == ValueArray([4, 9], 'm^2')).all())

        self.assertTrue((ValueArray([2, 3], 'GHz') * Value(3, 'ns')).dtype == np.float64)

        # isfinite
        self.assertTrue(np.isfinite(ValueArray([1, 2], 'GHz')).all())
        self.assertTrue((np.isfinite(ValueArray([1, float('nan')], 'GHz')) == np.array([True, False])).all())

    def testNegativePowers(self):
        self.assertEqual(str(units.Unit('1/s')), 's^-1')
        self.assertEqual(str(units.Unit('1/s^1/2')), 's^-1/2')

    def testTypeConversions(self):
        m = units.Unit('m')
        V = units.Unit('V')
        GHz = units.Unit('GHz')
        x1 = 1.0*m
        x2 = 5j*V
        a = np.arange(10)*1.0
        va = units.ValueArray(np.arange(10)*1.0, 'GHz')

        # Unit times number
        self.assertIsInstance(1.0*m, units.Value)
        self.assertIsInstance(1*m, units.Value)
        self.assertIsInstance(m*1.0, units.Value)
        self.assertIsInstance(m*1, units.Value)

        # Value times value or number
        self.assertIsInstance(x1*x1, units.Value)
        self.assertIsInstance(x1*5, units.Value)
        self.assertIsInstance(0*x1, units.Value)

        # Unit times complex
        self.assertIsInstance((1+1j)*V, units.Complex)
        self.assertIsInstance(V*(1+1j), units.Complex)

        # Value times Complex/complex
        self.assertIsInstance(x1*1j, units.Complex)
        self.assertIsInstance(1j*x1, units.Complex)
        self.assertIsInstance(x2*x1, units.Complex)
        self.assertIsInstance(x1*x2, units.Complex)

        # Unit/Value/ValueArray times array
        self.assertIsInstance(x1*a, units.ValueArray)
        self.assertIsInstance(x2*a, units.ValueArray)
        self.assertIsInstance(GHz*a, units.ValueArray)
        self.assertIsInstance(va*a, units.ValueArray)

        # Unit/Value/ValueArray times ValueArray
        self.assertIsInstance(x1*va, units.ValueArray)
        self.assertIsInstance(x2*va, units.ValueArray)
        self.assertIsInstance(GHz*va, units.ValueArray)
        self.assertIsInstance(va*va, units.ValueArray)

        # array times ?
        self.assertIsInstance(a*x1, units.ValueArray)
        self.assertIsInstance(a*x2, units.ValueArray)
        self.assertIsInstance(a*GHz, units.ValueArray)
        self.assertIsInstance(a*va, units.ValueArray)

        # ValueArray times ?
        self.assertIsInstance(va*x1, units.ValueArray)
        self.assertIsInstance(va*x2, units.ValueArray)
        self.assertIsInstance(va*GHz, units.ValueArray)
        self.assertIsInstance(va*a, units.ValueArray)

    def testComparison(self):
        s = units.Unit('s')
        ms = units.Unit('ms')
        kg = units.Unit('kg')
        self.assertTrue(1*s > 10*ms, '1*s > 10*ms')
        self.assertTrue(1*s >= 10*ms, '1*s >= 10*ms')
        self.assertTrue(1*s < 10000*ms, '1*s > 10000*ms')
        self.assertTrue(1*s <= 10000*ms, '1*s >= 10000*ms')
        self.assertTrue(10*ms < 1*s, '10*ms < 1*s')
        self.assertTrue(10*ms <= 1*s, '10*ms <= 1*s')
        self.assertTrue(10000*ms > 1*s, '10000*ms < 1*s')
        self.assertTrue(10000*ms >= 1*s, '10000*ms <= 1*s')
        with self.assertRaises(TypeError):
            nogood = 1*s > 1*kg

        self.assertFalse(1*s == 1*kg)
        self.assertTrue(0*s == 0)
        self.assertTrue(4*s > 0)
        with self.assertRaises(TypeError): _ = 4*s > 1

    def testComplex(self):
        V = units.Unit('V')

        self.assertTrue(1j*V != 1.0*V)
        self.assertTrue(1j*V == 1.0j*V)
        self.assertTrue(1.0*V == (1+0j)*V)
        with self.assertRaises(TypeError): _ = 1.0j*V < 2j*V

    def testDimensionless(self):
        ns = units.Unit('ns')
        GHz = units.Unit('GHz')

        self.assertTrue(isinstance((5*ns)*(5*GHz), float))
        self.assertTrue(hasattr((5*ns)*(5*GHz), 'inUnitsOf'))
        self.assertTrue(((5*ns)*(5*GHz)).isDimensionless())
        self.assertTrue((5*ns)*(5*GHz) < 50)
        self.assertTrue(isinstance(units.WithUnit(5.0, ''), units.DimensionlessFloat))

        self.assertTrue((5*ns*5j*GHz) == 25j)
        self.assertTrue((5*ns*5j*GHz).isDimensionless())

    def testAngle(self):
        rad = units.Unit('rad')
        self.assertTrue(rad.is_angle)
        self.assertTrue(rad.isAngle())
        x = units.Unit('rad*m/s')
        self.assertFalse(x.is_angle)
        
    def testInfNan(self):
        ms = units.Unit('ms')
        GHz = units.Unit('GHz')
        MHz = units.Unit('MHz')
        
        self.assertEquals(float('inf')*GHz, float('inf')*MHz)
        self.assertNotEqual(float('inf')*GHz, float('inf')*ms)
        self.assertNotEqual(float('inf')*GHz, -float('inf')*GHz)
        self.assertNotEqual(float('nan')*GHz, float('nan')*GHz)
        self.assertNotEqual(float('nan')*GHz, float('nan')*ms)
        
    def testPickling(self):
        ns = units.Unit('ns')
        GHz = units.Unit('GHz')
        blank = units.Unit('')

        def round_trip(obj):
            return cPickle.loads(cPickle.dumps(obj))
        self.assertEqual(round_trip(5*GHz), 5*GHz) # Value
        self.assertEqual(round_trip(GHz), GHz)     # Unit
        self.assertTrue((round_trip(np.arange(5)*ns) == np.arange(5)*ns).all()) # array
        self.assertEqual(round_trip(5*GHz*ns), 5)  # Dimensionless
        self.assertIsInstance(round_trip(3*blank), type(3*blank)) # Don't loose dimensionless type

    def testUnitCreation(self):
        self.assertIsInstance(units.Unit('test0', 1.0, units.hplanck/(2*units.e)), units.Unit)
        self.assertTrue((units.Unit('phi0')**2).isCompatible(units.Unit('phi0^2')))

    def testInUnitsOf(self):
        s = units.Unit('s')
        ms = units.Unit('ms')
        self.assertTrue((1*s).inUnitsOf(ms) == 1000*ms)
        self.assertTrue((1*s).inUnitsOf('ms') == 1000*ms)

    def testBaseUnitPowers(self):
        x = Value(1, 'ns^2')

        self.assertTrue(x.unit.base_unit == units.Unit('s^2'))
        self.assertTrue(x.inBaseUnits() == Value(1e-18, 's^2'))

    def testUnitPowers(self):
        self.assertTrue(units.Unit('ns')**2 == units.Unit('ns^2'))

    def test_array_priority(self):
        """numpy issue 6133

        DimensionlessX needs to support all arithmetic operations when the
        other side is an array.  Numpy's __array_priority__ machinery doesn't
        handle NotImplemented results correctly, so the higher priority element
        *must* be able to handle all operations.

        In numpy 1.9 this becomes more critical because numpy scalars like np.float64
        get converted to arrays before being passed to binary arithmetic operations.
        """
        x = np.float64(1)
        y = units.DimensionlessFloat(2)
        self.assertTrue(x < y)
        z = np.arange(5)
        self.assertTrue(((x<z) == [False, False, True, True, True]).all())

    def testNone(self):
        with self.assertRaises(Exception):
            units.Unit(None)
        with self.assertRaises(TypeError):
            None * units.Unit('MHz')

    def test_non_SI(self):
        units.addNonSI('count', True)
        x = 5 * units.Unit('kcount')
        self.assertTrue(x['count'] == 5000.0)
        self.assertTrue(x.inBaseUnits() == 5000.0*units.Unit('count'))
        self.assertTrue((x**2).unit == units.Unit('kcount^2'))

    def test_string_unit(self):
        ts = units.Unit('tshirt/s')
        self.assertEqual((1*ts)['tshirt/h'], 3600.0)
        self.assertEqual(str(ts), 'tshirt/s')

if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from labrad import util
from labrad.server import LabradServer, setting

from twisted.python import log
from twisted.internet.defer import inlineCallbacks

from datetime import datetime

DELAY_TIME = .25 # time to wait between measurements

class LoopServer(LabradServer):
    name = 'Test Looper'
    delayTime = DELAY_TIME
    testMode = True

    queries = [dict(server='manager',
                    setting='servers',
                    last=('', datetime.now()))]


    def initServer(self):
        self.alive = True
        self.readLoop().addErrback(log.err)

    def stopServer(self):
        self.alive = False

    @inlineCallbacks
    def readLoop(self):
        cxn = self.client
        while self.alive:
            yield cxn.refresh()
            for s in self.queries:
                server = s['server']
                setting = s['setting']
                try:
                    rslt = yield cxn[server][setting]()
                    s['last'] = str(rslt[0]), datetime.now()
                except Exception:
                    pass
            yield util.wakeupCall(self.delayTime)


    @setting(2, 'Queries', returns=['*s'])
    def list_queries(self, c):
        """A list of settings I query."""
        return [s['server'] + '.' + s['setting'] for s in self.queries]


    @setting(1, 'Get Readings', returns=['*(st)'])
    def get_readings(self, c):
        """Requests current readings."""
        return [s['last'] for s in self.queries]


    @setting(10, 'Delay Time', delay=['v[s]'], returns=['v[s]'])
    def delay_time(self, c, delay=None):
        if delay is not None:
            self.delayTime = delay
        return self.delayTime


__server__ = LoopServer()

if __name__ == '__main__':
    util.runServer(__server__)



# Copyright (C) 2008  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twisted.internet.defer import inlineCallbacks, returnValue
from labrad.support import SafeIterDict

class RegistryWrapperAsync(SafeIterDict):

    """Wrapper around the labrad registry.

    The LabRAD registry can be used like a python class or a list.
    Registry entries are mapped to valid python identifiers by
    replacing all characters except letters and digits with '_', and by
    prefixing with '_' if the first letter is a digit.
    Python identifiers are mapped to registry entries by removing all '_'
    at the beginning or end and by replacing '_' in the middle with ' '.
    This means that not all registry entries can be accessed as attributes.
    For registry entries that are not valid python names you should use
    attributes e.g. mywrapper['Key Name'].
    """

    MESSAGE_ID = 463462

    @classmethod
    @inlineCallbacks
    def create(cls, cxn, directory=''):
        """Create a registry wrapper and initialize it."""
        wrapper = cls(cxn, directory)
        yield wrapper._init()
        returnValue(wrapper)

    def __init__(self, cxn, directory=''):
        """Basic (synchronous) initialization."""
        self._cxn = cxn
        self._ctx = cxn.context()
        if not isinstance(directory, list):
            directory = [directory]
        if directory[0] != '':
            directory = [''] + directory
        self._dir = directory
        self._listen(True)

    @inlineCallbacks
    def _init(self):
        """Initialize this registry wrapper.

        We get the data for this directory and create wrappers for
        all subdirectories.  We also sign up for change notifications.
        """
        p = self._packet()
        p.cd(self._dir, True)
        p.dir()
        ans = yield p.send()
        dirs, keys = ans['dir']
        _dict = SafeIterDict()
        for d in dirs:
            _dict[d] = self._wrapSubdir(d)
        p = self._packet()
        p.notify_on_change(self.MESSAGE_ID, True)
        for k in keys:
            p.get(k, key=k)
        ans = yield p.send()
        for d in dirs:
            _dict[d] = yield _dict[d]
        for k in keys:
            _dict[k] = ans[k]
        self.update(_dict)
        print 'done initing!'
        print 'data:', self

    def __del__(self):
        """Remove the listener when we are garbage collected."""
        self._listen(False)

    def _listen(self, listen):
        """Add or remove a listener for update messages."""
        #def print_(*args):
        #    print 'message:', args
        #    #print self._cxn._cxn.listeners
        #self._cxn._cxn.addListener(print_)
        if listen:
            func = self._cxn._cxn.addListener
        else:
            func = self._cxn._cxn.removeListener
        source = self._cxn.registry.ID # doesn't seem to work
        func(self._messageReceived, context=self._ctx, ID=self.MESSAGE_ID)

    def _wrapSubdir(self, dir):
        """Create a registry wrapper of a subdirectory."""
        return self.__class__.create(self._cxn, self._dir + [dir])

    def _getKey(self, key):
        """Get a key from this directory."""
        return self._cxn.registry.get(key, context=self._ctx)

    def _packet(self):
        """Create a packet to the registry."""
        return self._cxn.registry.packet(context=self._ctx)

    @inlineCallbacks
    def _messageReceived(self, c, data):
        """Handle update messages from the registry."""
        name, isDir, addOrChange = data
        print 'update message:', data
        try:
            if addOrChange:
                if isDir:
                    val = yield self._wrapSubdir(name)
                else:
                    val = yield self._getKey(name)
                self[name] = val
            else:
                del self[name]
            print 'done updating!'
            print 'data:', self
        except:
            print 'Error while updating Registry Wrapper for', self._dir
            from twisted.python.failure import Failure
            print Failure().getBriefTraceback()



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.thread

Methods to provide a blocking, threaded interface to the
asynchronous twisted backend.
"""

import threading

from twisted.internet import reactor, defer
from twisted.python import threadable
from twisted.python.failure import Failure

threadable.init(1)

class ReactorThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, name='ReactorThread')
        self.daemon = True
    def run(self):
        reactor.run(installSignalHandlers=0)

_reactorThread = None

def startReactor():
    # check to see whether the reactor is already running
    # this ensures that when synchronous code is called from an asynchronous
    # program using deferToThread we don't try to restart the reactor
    if reactor.running:
        return
    global _reactorThread
    if not _reactorThread or not _reactorThread.isAlive():
        _reactorThread = ReactorThread()
        _reactorThread.start()

def stopReactor():
    reactor.callFromThread(reactor.stop)




# This module implements a LRU cache.  There are many modules like it,
# but this one is mine.
#
# The basic idea of a cache is that you have some function f(x) that is
# expensive (in CPU or IO) to compute, and for which we often evaluate for
# the same x many times.  If f(x) depends only on x and has no side effects,
# you can just remember (memoize) the result after the first call, and subsequent
# calls just return the cached result.
#
# There are a couple of problems with that.  First, a generic problem: caches
# need a mechanism to discard entries or they grow without bound.  In our
# case, the cache values may be relatively large arrays: 100 KB - 1 MB.  If
# we store 1000 entries, that can take up to 1 GB of memory.  So real caches
# eventually need a mechanism to discard entries.  The most common algorithms
# are least-frequently-used and least-recently-used.  When the cache reaches
# a fixed size, the indicated elements are discarded.  Often it is computationally
# easier to find some approximation of the least-recently-used element.
#
# The second problem is that in our case 'x' itself may be quite large (a numpy
# array of similar size to f(x).  Locating an item x by value may be excessively
# expensive.  numpy arrays are not hashable, and not immutable, complicating
# matters further.  Therefore, it is the client's responsibility to calculate a
# unique key that represents the data.  A common pattern will be to require that
# x be an array of equally spaced elements, which can be uniquely described by the
# the first and last elements, and the length.

from collections import namedtuple
from functools import update_wrapper
from threading import RLock

import numpy as np

_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])

class _HashedSeq(list):
    __slots__ = 'hashvalue'

    def __init__(self, tup, hash=hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue

def make_key(args, kwds, typed=False,
             kwd_mark=(object(),),
             fasttypes=[int, str, frozenset, type(None)]):
    """Make a cache key from optionally typed positional and keyword arguments.

    Raises TypeError if any items in args or values in kwds are not hashable.
    """
    key = tuple(args)
    if kwds:
        sorted_items = sorted(kwds.items())
        key += kwd_mark
        for item in sorted_items:
            key += item
    if typed:
        key += tuple(type(v) for v in args)
        if kwds:
            key += tuple(type(v) for k, v in sorted_items)
    elif len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    return _HashedSeq(key)

class fast_str(str):
    """String class with a hash function that is fast for very long strings.

    Note that str(a) and fast_str(a) will compare equal but
    have different hash values.  Do not use unless you understand the
    ramifications of that.
    """
    def __hash__(self):
        if len(self) > 64:
            return str.__hash__(self[0:32:2] + self[-32::2])
        else:
            return str.__hash__(self)

def keyfunc_ndarray(a):
    """Key function for a numpy array.

    Uses the string representation of the raw data,
    and uses the fast_str hash function for maximum performance.
    """
    a = np.asarray(a)
    return (a.shape, a.dtype, fast_str(a.tostring()))


class LRUCache(object):
    """Least-recently-used cache dictionary.

    Get cached results by calling the 'get' method, passing a key to lookup
    and a function to be called with specified positional and keyword args
    to compute the result if the key is not found. The cache will store up
    to *maxsize* elements, tossing out the least-recently used element
    when the maximum size is reached.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used

    Based on a backport of functools.lru_cache from python 3.3+:
    http://code.activestate.com/recipes/578078-py26-and-py30-backport-of-python-33s-lru-cache/

    Implementation details:
    The LRU cache consists of a dictionary (self.cache) and a circular
    doubly-linked list. The dictionary contains a mapping from keys to
    linked-list elements, and each linked list element contains 4 entries:
    the list prev- and next-pointers, the key, and the cached result value.
    """

    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3 # names for the link fields

    def __init__(self, maxsize=128):
        self.maxsize = maxsize

        # depending on maxsize, choose the implementation of 'get'
        # based on whether or not we need caching and size limitation
        if maxsize == 0:
            self.get = self._get_nocache
        elif maxsize is None:
            self.get = self._get_nomax
        else:
            self.get = self._get

        self.hits = self.misses = 0
        self.cache = {}
        self.lock = RLock() # because linkedlist updates aren't threadsafe
        root = self.root = [] # root of the circular doubly-linked list
        root[:] = [root, root, None, None] # link the list to itself

    def _get_nocache(self, key, fn, *args, **kwds):
        """no caching, just do a statistics update after a successful call"""
        result = fn(*args, **kwds)
        self.misses += 1
        return result

    def _get_nomax(self, key, fn, *args, **kwds):
        """simple caching without ordering or size limit"""
        result = self.cache.get(key, self.root) # root used here as a unique not-found sentinel
        if result is not self.root:
            self.hits += 1
            return result
        result = fn(*args, **kwds)
        self.cache[key] = result
        self.misses += 1
        return result

    def _get(self, key, fn, *args, **kwds):
        """size-limited caching that tracks accesses by recency"""
        with self.lock:
            link = self.cache.get(key)
            if link is not None:
                # record recent use of the key by moving it to the front of the list
                root = self.root
                link_prev, link_next, key, result = link
                link_prev[self.NEXT] = link_next
                link_next[self.PREV] = link_prev
                last = root[self.PREV]
                last[self.NEXT] = root[self.PREV] = link
                link[self.PREV] = last
                link[self.NEXT] = root
                self.hits += 1
                return result
        result = fn(*args, **kwds)
        with self.lock:
            root = self.root
            if key in self.cache:
                # getting here means that this same key was added to the
                # cache while the lock was released.  since the link
                # update is already done, we need only return the
                # computed result and update the count of misses.
                pass
            elif len(self.cache) >= self.maxsize:
                # use the old root to store the new key and result
                oldroot = root
                oldroot[self.KEY] = key
                oldroot[self.RESULT] = result
                # empty the oldest link and make it the new root
                root = self.root = oldroot[self.NEXT]
                oldkey = root[self.KEY]
                oldvalue = root[self.RESULT]
                root[self.KEY] = root[self.RESULT] = None
                # now update the cache dictionary for the new links
                del self.cache[oldkey]
                self.cache[key] = oldroot
            else:
                # put result in a new link at the front of the list
                last = root[self.PREV]
                link = [last, root, key, result]
                last[self.NEXT] = root[self.PREV] = self.cache[key] = link
            self.misses += 1
        return result

    def info(self):
        """Report cache statistics"""
        with self.lock:
            return _CacheInfo(self.hits, self.misses, self.maxsize, len(self.cache))

    def clear(self):
        """Clear the cache and cache statistics"""
        with self.lock:
            self.cache.clear()
            root = self.root
            root[:] = [root, root, None, None]
            self.hits = self.misses = 0

    def evict(self, key):
        """Remove the given key from the cache."""
        with self.lock:
            if key in self.cache:
                link = self.cache[key]
                # remove item from the lru linked list
                link[self.PREV][self.NEXT] = link[self.NEXT]
                link[self.NEXT][self.PREV] = link[self.PREV]
                # remove item from the cache
                del self.cache[key]

def lru_cache(maxsize=128, typed=False):
    """Least-recently-used cache decorator.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    If *typed* is True, arguments of different types will be cached separately.
    For example, f(3.0) and f(3) will be treated as distinct calls with
    distinct results.

    Uses the make_key function to create a hash key for arguments passed
    to the decorated function. Hence, arguments to the cached function
    must be hashable, or the call will fail with a TypeError.

    View the cache statistics named tuple (hits, misses, maxsize, currsize) with
    f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.

    Based on a backport of functools.lru_cache from python 3.3+:
    http://code.activestate.com/recipes/578078-py26-and-py30-backport-of-python-33s-lru-cache/
    """

    def decorating_function(user_function):

        cache = LRUCache(maxsize)

        def wrapper(*args, **kwds):
            key = make_key(args, kwds)
            return cache.get(key, user_function, *args, **kwds)

        def cache_info():
            """Report cache statistics"""
            return cache.info()

        def cache_clear():
            """Clear the cache and cache statistics"""
            cache.clear()

        def cache_evict(key):
            """Remove the given key from the cache."""
            cache.evict(key)

        wrapper.__wrapped__ = user_function
        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        wrapper.cache_evict = cache_evict
        return update_wrapper(wrapper, user_function)

    return decorating_function



import collections
import contextlib
import os
import shutil
import subprocess
import sys
import tempfile
import time

import pytest

import labrad
from labrad import crypto


@contextlib.contextmanager
def temp_tls_dirs():
    """Context manager to create temporary dirs for labrad TLS certs and keys.

    Also overrides the path labrad.crypto.CERTS_PATH where we look for trusted
    certs to correspond to the temporary directory. If we then also point the
    manager to this directory to store its generated self-signed certificates,
    then we will automatically trust those certificates.

    Yields (string, string):
        A tuple of paths for the TLS certs and keys, respectively.
    """
    user_dir = os.path.expanduser('~')
    cert_path = tempfile.mkdtemp(prefix='.labrad-test-certs', dir=user_dir)
    key_path = tempfile.mkdtemp(prefix='.labrad-test-keys', dir=user_dir)
    old_cert_path = crypto.CERTS_PATH
    crypto.CERTS_PATH = cert_path
    try:
        yield cert_path, key_path
    finally:
        crypto.CERTS_PATH = old_cert_path
        shutil.rmtree(cert_path)
        shutil.rmtree(key_path)


# Manager Info class returned by the run_manager cont
ManagerInfo = collections.namedtuple('ManagerInfo', ['port', 'tls_port', 'password'])


@contextlib.contextmanager
def run_manager(tls_required, port=7778, tls_port=7779, startup_timeout=20):
    """Context manager to run the labrad manager in a subprocess.

    Will attempt to connect to the manager and fail if we cannot do so within
    the specified timeout. This ensures that the manager is actually running
    and listening for connections before we yield to execute the body of the
    with statement.

    Args:
        tls_required (boolean): Whether the manager should require TLS.
        port (int): The port on which to listen for upgradeable connections
            that start unencrypted and then use STARTTLS to secure the
            connection.
        tls_port (int): The port on which to listen for TLS connections that
            are encrpyted from the start.
        startup_timeout (float): Maximum number of seconds to allow for the
            manager to start before we fail.

    Yields (ManagerInfo):
        Info about the running manager.
    """
    with temp_tls_dirs() as (cert_path, key_path):
        password = 'DummyPassword'
        manager = subprocess.Popen([
                'labrad',
                '--password={}'.format(password),
                '--port={}'.format(port),
                '--tls-port={}'.format(tls_port),
                '--tls-required={}'.format(tls_required),
                '--tls-required-localhost={}'.format(tls_required),
                '--tls-cert-path={}'.format(cert_path),
                '--tls-key-path={}'.format(key_path)])
        try:
            start = time.time()
            while True:
                try:
                    labrad.connect(port=tls_port, tls_mode='on',
                                   password=password)
                except Exception, e:
                    last_error = e
                else:
                    break
                elapsed = time.time() - start
                if elapsed > startup_timeout:
                    raise Exception('labrad failed to start within {} seconds. '
                                    'last_error={}'
                                    .format(startup_timeout, last_error))
                time.sleep(0.5)
            yield ManagerInfo(port, tls_port, password)
        finally:
            manager.kill()


# Test that we can establish encrypted TLS connections to the manager

def test_connect_with_starttls():
    with run_manager(tls_required=True) as m:
        with labrad.connect(port=m.port, tls_mode='starttls-force',
                            password=m.password) as cxn:
            pass

def test_connect_with_optional_starttls():
    with run_manager(tls_required=False) as m:
        with labrad.connect(port=m.port, tls_mode='off',
                            password=m.password) as cxn:
            pass

def test_connect_with_tls():
    with run_manager(tls_required=True) as m:
        with labrad.connect(port=m.tls_port, tls_mode='on',
                            password=m.password) as cxn:
            pass


# Test that connecting to the manager fails if the client fails to
# use TLS when the manager expects it.

def test_expect_starttls_use_off():
    with run_manager(tls_required=True) as m:
        with pytest.raises(Exception):
            with labrad.connect(port=m.port, tls_mode='off',
                                password=m.password) as cxn:
                pass

def test_expect_tls_use_off():
    with run_manager(tls_required=True) as m:
        with pytest.raises(Exception):
            with labrad.connect(port=m.tls_port, tls_mode='off',
                                password=m.password) as cxn:
                pass

def test_expect_tls_use_starttls():
    with run_manager(tls_required=True) as m:
        with pytest.raises(Exception):
            with labrad.connect(port=m.tls_port, tls_mode='off',
                                password=m.password) as cxn:
                pass


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import unittest

import labrad
from labrad import types as T
from labrad.servers.test_server import TestServer
from labrad.util import syncRunServer

TEST_STR = 'this is a test, this is only a test'

class ClientTests(unittest.TestCase):
    def run(self, result=None):
        """Override the TestCase run method to launch the test server.

        This seems to be the cleanest solution for effectively using a
        context manager as part of a test fixture.  This method then calls
        the usual test fixture setUp and tearDown within this context.
        """
        with syncRunServer(TestServer()):
            super(ClientTests, self).run(result)

    def setUp(self):
        self.cxn = labrad.connect()

    def tearDown(self):
        self.cxn.disconnect()

    def _get_manager(self):
        self.assertTrue(hasattr(self.cxn, 'manager'))
        return self.cxn.manager

    def _get_tester(self):
        self.assertTrue(hasattr(self.cxn, 'python_test_server'))
        return self.cxn.python_test_server

    def testConnection(self):
        servers = self.cxn.servers
        self.assertTrue(len(servers.keys()) > 0)
        self.assertTrue('manager' in servers)
        self.assertTrue('python_test_server' in servers)

        self._get_manager()
        self._get_tester()

        self._get_manager()
        self._get_tester()

    def testServer(self):
        pts = self._get_tester()

        # make sure we can access the setting by both allowed methods
        self.assertTrue(hasattr(pts, 'echo'))
        self.assertTrue('echo' in pts.settings)

        # single setting, named explicitly
        resp = pts.echo(TEST_STR)
        self.assertEqual(resp, TEST_STR)

        resp = pts.echo(T.Value(15.0, 's'))
        self.assertEqual(resp['s'], 15.0)
        self.assertEqual(resp.unit.name, 's')

        # single setting with the name looked up
        resp = pts.settings['echo']([1, 2, 3, 4])
        self.assertEqual(len(resp), 4)

        # single setting with delayed response
        resp = pts.echo.future(TEST_STR)
        resp = resp.result()
        self.assertEqual(resp, TEST_STR)

        # allow calling with wait=False for backwards compatibility
        resp = pts.echo(TEST_STR, wait=False)
        resp = resp.result()
        self.assertEqual(resp, TEST_STR)

    def testCompoundPacket(self):
        pts = self._get_tester()

        self.assertTrue(hasattr(pts, 'packet'))

        # make a simple packet and check the various methods of getting
        # data out of it
        pkt = pts.packet()
        resp = pkt.echo(1).echo(2).settings['echo'](3).send()
        self.assertTrue(hasattr(resp, 'echo'))
        self.assertEqual(len(resp.echo), 3)
        self.assertTrue('echo' in resp.settings)
        self.assertEqual(len(resp.settings['echo']), 3)
        self.assertEqual(len(resp['echo']), 3)

        # test using keys to refer to parts of a packet
        pkt2 = pts.packet()
        resp = pkt2.echo(1L, key='one')\
                   .echo_delay(T.Value(0.1, 's'))\
                   .delayed_echo('blah', key='two')\
                   .send()
        self.assertTrue(hasattr(resp, 'one'))
        self.assertTrue('one' in resp.settings)
        self.assertEqual(resp['one'], 1)
        self.assertFalse(hasattr(resp, 'echo_word'))
        self.assertEqual(resp.two, 'blah')

        # test packet mutation by key
        pkt2['two'] = TEST_STR
        resp = pkt2.send()
        self.assertEqual(resp.two, TEST_STR)

        # send packet asynchronously
        resp = pkt2.send_future()
        resp = resp.result()
        self.assertEqual(resp.two, TEST_STR)

        # allow sending with wait=False for backwards compatibility
        resp = pkt2.send(wait=False)
        resp = resp.result()
        self.assertEqual(resp.two, TEST_STR)

    def testTupleKeys(self):
        # allow the use of tuples as packet keys
        pts = self._get_tester()

        p = pts.packet()
        p.echo(1, key='a')
        p.echo(2, key=(1, 2))
        r = repr(p)
        s = str(p)
        resp = p.send()
        self.assertEqual(resp.a, 1)
        self.assertEqual(resp['a'], 1)
        self.assertEqual(resp[(1, 2)], 2)

    def testExceptions(self):
        pts = self._get_tester()

        pts.echo_delay(T.Value(0.1, 's'))
        self.assertRaises(Exception, pts.exc_in_handler)
        self.assertRaises(Exception, pts.exc_in_subfunction)
        self.assertRaises(Exception, pts.exc_in_deferred)
        self.assertRaises(Exception, pts.exc_in_errback)
        self.assertRaises(Exception, pts.exc_in_inlinecallback)

    def testContextWrappers(self):
        cxn1 = self.cxn()
        cxn2 = self.cxn()

        pts1 = cxn1.python_test_server
        pts2 = cxn2.python_test_server
        pts3 = cxn1.python_test_server()
        pts4 = pts2()

        pts1.set('1', 1)
        pts2.set('2', 2)
        pts3.set('3', 3)
        pts4.set('4', 4)

        self.assertEqual(pts1.keys(), ['1'])
        self.assertEqual(pts2.keys(), ['2'])
        self.assertEqual(pts3.keys(), ['3'])
        self.assertEqual(pts4.keys(), ['4'])

    def test_server_calls(self):
        """Make sure that server settings can call other settings directly.
        """
        cxn = self.cxn()

        ts = cxn.python_test_server

        ts.set_reversed("bar", "foo")
        self.assertEqual(ts.get("foo"), "bar")

        ts.echo_delay(T.Value(0.1, 's'))
        self.assertEquals(ts.delayed_echo_wrapper(1), 1)

if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2007  Matthew Neeley
# Copyright (C) 2014, Daniel Sank, Evan Jeffrey
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
LabRAD <-> Python data conversion

All data in LabRAD packets is strictly typed, while python data
is dynamically typed.  But we can do a pretty good job converting
automatically back and forth between the two, and when we have
information about what types are allowed (for example when sending
to a server setting or returning from a setting when the accepted
and return types have been registered), we can do even better.
"""

from __future__ import absolute_import

import collections
import datetime
import itertools
import re
import struct
import sys
import time
from types import InstanceType
import warnings

import labrad.units as U
from labrad.units import Value, Complex

import numpy as np


SYSTEM_BYTE_ORDER = '<' if sys.byteorder == 'little' else '>'


# helper classes

class Singleton(object):
    """Base class for singleton types.

    A singleton is a type with only one instance.  The first instance
    created gets stored in the class attribute _inst, and this same
    instance is returned thereafter.
    """
    def __new__(cls, *a, **kw):
        inst = getattr(cls, '_inst', None)
        if inst is None:
            inst = super(Singleton, cls).__new__(cls, *a, **kw)
            cls._inst = inst
        return inst


# a registry of parsing functions, keyed by type tag
_parsers = {} # type tag -> parse function


class RegisterParser(type):
    """A metaclass for LabRAD types that have parsers.

    When each class is created, this function gets called and
    we register the __parse__ method of the class according
    to the tag that the class specifies.
    """
    def __new__(cls, name, bases, dict):
        c = type.__new__(cls, name, bases, dict)
        if 'tag' in dict:
            _parsers[dict['tag']] = c.__parse__
        return c


class Buffer(object):
    """Consume strings without memory reallocation to improve performance."""

    def __init__(self, s):
        if isinstance(s, Buffer):
            self.s = s.s
            self.ofs = s.ofs
        else:
            self.s = s
            self.ofs = 0

    def get(self, i=1):
        """Get i characters from the buffer and increment pointer."""
        temp = self.s[self.ofs:self.ofs+i]
        self.ofs += i
        return temp

    def skip(self, i=1):
        """Increment pointer."""
        self.ofs += i

    def __len__(self):
        """Length of buffer which remains to be consumed."""
        return len(self.s) - self.ofs

    def __str__(self):
        """Returns unconsumed part of buffer."""
        return self.s[self.ofs:]

    def __getitem__(self, key):
        """Get part of unconsumed buffer indexed by key.

        If the current pointer is N and we call __getitem__(m), we get the
        character with absolute index N+m.

        This method reallocates the unconsumed part of the buffer to self.s
        and resets the offset pointer. That means we're doing string allocation,
        so don't use this for performance code.
        """
        self.s = self.s[self.ofs:]
        self.ofs = 0
        return self.s[key]

    def strip(self, chars):
        """Strip characters from the string and reallocate.

        Do not use this in performance code.
        """
        self.s = self.s[self.ofs:].strip(chars)
        self.ofs = 0
        return self

    def index(self, char):
        """Find index of char in uncomsumed part of buffer."""
        return self.s.index(char, self.ofs)


# typetag parsing

def parseTypeTag(s):
    """Parse a type tag into a LabRAD type object."""
    try:
        if isinstance(s, LRType):
            return s
        s = stripComments(s)
        ## this is a workaround for a bug in the manager
        ## What bug? This needs to be explained.
        if s == '' or s[:1] == '_':
            return LRNone()
        # Comments were already stripped, so why does the next line exist? DTS
        s = Buffer(stripComments(s))
        types = []
        while len(s):
            types.append(parseSingleType(s))
        if len(types) == 0:
            return LRNone()
        elif len(types) == 1:
            return types[0]
        else:
            return LRCluster(*types)
    except Exception:
        print 'failed to parse:', s
        raise

WHITESPACE = ' ,\t'

def parseSingleType(s):
    """Parse a single type at the beginning of a type string.

    s must be an instance of Buffer.
    """
    s.strip(WHITESPACE)
    # Consume one character to determine parsing function and then act that
    # parsing function of the remaining Buffer.
    t = _parsers[s.get()](s)
    s.strip(WHITESPACE)
    return t

COMMENTS = re.compile('\{[^\{\}]*\}')

def stripComments(s):
    """Remove comments from a type tag.

    Inline comments are delimited by curly brackets {} and may not be nested.
    In addition, anything after a colon : is considered a comment.
    """
    return COMMENTS.sub('', s).split(':')[0].strip()

def parseNumber(s):
    """Parse an integer at the beginning of a string."""
    s.strip(WHITESPACE)
    n = d = 0
    while len(s) and s[:1] in '0123456789':
        n = 10*n + int(s.get())
        d += 1
    if d == 0:
        n = None # no digits
    return n

def parseUnits(s):
    """Parse units from a typestring."""
    s = s.strip(WHITESPACE)
    if s[:1] != '[':
        return None
    length = s.index(']') + 1
    return s.get(length)[1:-1]


# a registry of types and type functions that can determine
# the LabRAD type of python data
_types = {} # python class -> LabRAD type

def registerType(cls, t):
    """Register the LabRAD type for a python class.

    This is used in a case where the LabRAD type is known
    immediately from the python class.
    """
    classes = cls if isinstance(cls, tuple) else (cls,)
    t = parseTypeTag(t) if isinstance(t, str) else t
    for cls in classes:
        _types[cls] = t

_typeFuncs = {}

def registerTypeFunc(cls, typeFunc):
    """Register a LabRAD type function for a python class.

    The type function takes a python object of class cls and
    returns an appropriate LabRAD type object for it.
    """
    classes = cls if isinstance(cls, tuple) else (cls,)
    for cls in classes:
        _typeFuncs[cls] = typeFunc

def getType(obj):
    """
    Get LabRAD type of python data.

    We try the following ways of finding of finding a suitible type, in order:

    obj.__lrtype__()
    _types[type(obj)]
    _typeFuncs[type(obj)](obj)
    _typeFuncs[superclasses of obj](obj)
    """
    if hasattr(obj, '__lrtype__'):
        return obj.__lrtype__()

    t = type(obj)

    if t == FlatData:
        return parseTypeTag(obj.tag)

    # handle classic classes
    if t == InstanceType:
        t = obj.__class__

    # check if we know this type
    if t in _types:
        return _types[t]

    # check if we have a type function for this type
    if t in _typeFuncs:
        return _typeFuncs[t](obj)
    else:
        # check if we have a type function for a superclass
        for cls in _typeFuncs:
            if issubclass(t, cls):
                return _typeFuncs[cls](obj)
    raise TypeError("No LabRAD type for: %r (type %s)" % (obj, type(obj)))


# flattening and unflattening

def unflatten(s, t, endianness='>'):
    """Unflatten labrad data into python data, given a prototype t.

    s - string or Buffer: data to be unflattened.
    t - string or LRType subclass: the prototype. If string this is a LabRAD
        type tag.

    The prototype can be a labrad type tag or a prototype object
    created the parseTypeTag function.  At present, the default
    unflatteners are called at each stage, according to t.
    """
    if isinstance(t, str):
        t = parseTypeTag(t)
    if isinstance(s, str):
        s = Buffer(s)
    return t.__unflatten__(s, endianness)


class FlatData(collections.namedtuple('FlatDataBase', ['bytes', 'tag', 'endianness'])):
    """FlatData represents unflattened data.

    Attributes:

        bytes:      The data as intended to be transmitted over the wire
        tag:        The LabRAD data type in the form of an LRType object.
        endianness: The endiannes of the connection. '>' for big endian
                    and '<' for little endian.  Little endian is deprecated.
    """
    __slots__ = ()
    def unflatten(self, endianness='>'):
        return unflatten(self.bytes, self.tag, self.endianness)


def flatten(obj, types=None, endianness='>'):
    """Flatten python data into labrad data.

    Returns a tuple containing the flattened string and the type object
    representing the LabRAD data type to which obj was flattened.

    Flatten returns a tuple of (flattened string, type object).  The
    type object can be converted into a type tag by calling str(typeobj).

    We first try obj.__lrflatten__(endianness). If that fails we go to the list
    of suggested types in order until one of them works. If none of them work,
    we raise FlatteningError.

    XXX This is worthless without a specification of __lrflatten__'s
    return signature.

    Suggested types can be provided in the form of type tags (strings) or type
    objects as created by parseTypeTag.

    If not types are provided, we first check to see if the object has
    an __lrflatten__ method, in which case it is used. Then, we check
    the registry of flattening functions to see whether one exists for the
    object's type, or one of its superclasses.

    If no types are provided, we first check to see if the object has
    an __lrflatten__ method, in which case it will be called.  Then we
    check the registry of flattening functions, to see whether one exists
    for the object type, or a superclass.
    """
    if hasattr(obj, '__lrflatten__'):
        s, t = obj.__lrflatten__(endianness)
        return FlatData(s, t, endianness)

    if types is None:
        types = []
    if not isinstance(types, list):
        types = [types]
    types = [parseTypeTag(t) for t in types]
    t = getType(obj)
    if not len(types):
        # if there are no type suggestions, just try to
        # flatten to the default type
        return t.flatten(obj, endianness)

    # check the list of allowed types for one compatible
    # with obj's type
    foundCompatibleType = False
    for tt in types:
        # If obj has a type which is more specific than one of the
        # allowed types, use obj's type. This covers cases where
        # an allowed type is 'v' and obj's type is 'v[Hz]'.
        if t <= tt:
            foundCompatibleType = True
            break
        # If one of the required types is more specific than obj's
        # type, use that. This covers cases where we're trying to
        # flatten something like ([],[1,2]) and an allowed type is
        # eg. (*v*i).
        elif tt <= t:
            t = tt
            foundCompatibleType = True
            break
    if foundCompatibleType:
        try:
            return t.flatten(obj, endianness)
        except Exception, e:
            raise FlatteningError(obj, t)

    # Since we haven't found anything compatible, just try to
    # flatten to any of the suggested types. This covers cases
    # such as an allowed type 'v[]' and obj=4. In this case, the
    # class representing 'v[]', namely LRValue(''), will handle
    # coercion of obj to the appropriate type.
    for t in types:
        try:
            return t.flatten(obj, endianness)
        except Exception:
            pass
        else:
            break

    raise FlatteningError(obj, types)


# Evaluation functions

def evalLRData(s):
    """Evaluate labrad data from string form produced by reprLRData.

    DEPRECATED

    The format produced by reprLRData has been deprecated, and this function
    is kept solely for backwards compatibility to load data stored with the
    old format.
    """

    # Value and Complex constructors previously allowed the unit parameter
    # to be None, and old data may have been saved that way. These shims
    # convert the unit to '' instead and then call the real constructors.
    def _Value(x, unit=''):
        if unit is None:
            unit = ''
        return U.Value(x, unit)

    def _Complex(x, unit=''):
        if unit is None:
            unit = ''
        return U.Complex(x, unit)

    globs = globals()
    globs['Value'] = _Value
    globs['Complex'] = _Complex
    globs['nan'] = float('nan')
    globs['inf'] = float('inf')
    return eval(s, globs)

def reprLRData(s):
    """Make a string repr of labrad data that can be parsed with evalLRData.

    DEPRECATED

    This was an attempt to create a human-readable string representation of
    labrad data suitable for storing small amounts of data and reading it
    in later. It was used in early versions of the data vault, and is kept
    here for backwards-compatibility only. This should not be used for new
    applications; instead, just use the standard wire-format for serialization.
    """
    return repr(s)


# LabRAD type classes

class LRType(object):
    """Base class of all LabRAD type objects.

    These type classes manage parsing and creation of type tags,
    provide default unflatteners, and also provide methods to test
    type equality and compatibility.
    """

    __metaclass__ = RegisterParser

    width = 0
    isFixedWidth = True

    def __str__(self):
        """Convert to a type tag string format.

        All type object should return a LabRAD-compatible representation
        of themselves when __str__ is called, so that a type tag can
        be created from a type object simply by calling str(typeobj).
        """
        return self.tag

    def __repr__(self):
        """Convert to a verbose string representation.

        This string representation should be valid python code that could
        be evaluated to create the corresponding type object, assuming
        the names from the type module have been imported.
        """
        return self.__class__.__name__ + '()'

    @classmethod
    def __parse__(cls, s):
        """
        Parse type tag into appropriate python type objects.

        Returns the type object.

        The parse function takes the remaining string (after the tag for
        this type has been removed).  The function may need to consume
        some more of the string to determine the type.
        """
        return cls()

    @classmethod
    def __lrtype__(cls, obj):
        """
        Return the LabRAD type object for a python object.
        """
        return cls()

    def __cmp__(self, other):
        raise RuntimeError("Unreachable __cmp__ in LRType")

    def __eq__(self, other):
        """Test whether this type is equal to another.

        By default, we just check whether the types are the same.
        """
        return type(self) == type(other)

    def __le__(self, other):
        """
        Test whether this type is equal to on more specific than another.

        By default, just check for equality, or that the other type is LRAny.

        This operator has a very specific meaning which is easy to
        misunderstand. For two types, X and Y, X<=Y if and only if a setting
        advertising 'Y' will happily accept a LaBRAD record with data of type
        'X'. The simplest example of this is a setting accepting '?'. Such a
        setting will happily accept 'i', 'b', or anything else. Therefore, we
        have eg. LRInt()<=LRAny. A more complex example is 'v[Hz]' <= 'v'.
        """
        return type(self) == type(other) or type(other) == LRAny

    def isFullySpecified(self):
        return True

    def __width__(self, s, endianness):
        """
        Skip n bytes of tag buffer where n is width of this type.

        Returns number of bytes skipped.
        """
        s.skip(self.width)
        return self.width

    def __unflatten__(self, s, endianness):
        """Unflatten data from a string representation.

        Each LabRAD type object implements a default unflattener that
        creates python data from the LabRAD data string.  The unflattener
        returns a tuple of (python object, rest of string).
        """
        raise NotImplementedError

    def flatten(self, data, endianness):
        """Flatten an object to this Labrad type.

        This method deals with the case where incoming data has already been
        flattened. In that case, we check that the FlatData's type matches,
        otherwise for not-yet-flattened data we defer the real work to the
        __flatten__ method.

        Args:
            data (object or FlatData): If a FlatData is given, we just check
                that its type matches our type type and then return the already-
                flattened data. Otherwise, we call __flatten__.
            endianness (str): '<' for little- or '>' for big-endian.

        Returns:
            FlatData: The flattened data consisting of data and type tag.
        """
        if isinstance(data, FlatData):
            if not parseTypeTag(data.tag) <= self:
                raise FlatteningError(data, self)
            # TODO: @ejeffrey This is not an error, and we should
            # instead unflatten + flatten the data, possibly with a
            # warning, but I want to treat it as an error until this
            # is all working.
            if data.endianness != endianness:
                raise RuntimeError("Flattened data provided with wrong endianness")
            return data
        s, t = self.__flatten__(data, endianness)
        return FlatData(s, t, endianness)

    def __flatten__(self, data, endianness):
        """
        Flatten python data to this LabRAD type.

        Returns a tuple of (flattened string, LabRAD type object).

        Each LabRAD type object implements a default unflattener that
        creates python data from the LabRAD data string.  The unflattener
        returns a tuple of (flattened string, labrad type).  Note that
        other unflattener functions can be registered later to override
        this default behavior and unflatten to different python types.
        """
        raise NotImplementedError


class LRAny(LRType, Singleton):
    """A placeholder for any single LabRAD type."""
    tag = '?'

    def isFullySpecified(self):
        return False


class LRNone(LRType, Singleton):
    """An empty piece of LabRAD data."""
    tag = '_'

    def __unflatten__(self, s, endianness):
        return None

    def __flatten__(self, data, endianness):
        if data is not None:
            raise FlatteningError(data, self)
        return '', self

registerType(type(None), LRNone())
# register LRNone parser for the empty string as well
_parsers[''] = LRNone.__parse__


class LRBool(LRType, Singleton):
    """A simple boolean."""
    tag = 'b'
    width = 1

    def __unflatten__(self, s, endianness):
        return bool(ord(s.get(1)))

    def __flatten__(self, b, endianness):
        if not isinstance(b, (bool, np.bool8)):
            raise FlatteningError(b, self)
        return chr(b), self

registerType(bool, LRBool())
registerType(np.bool8, LRBool())

class LRInt(LRType, Singleton):
    """A signed 32-bit integer."""
    tag = 'i'
    width = 4
    def __unflatten__(self, s, endianness):
        return struct.unpack(endianness + 'i', s.get(4))[0]

    def __flatten__(self, n, endianness):
        if not isinstance(n, (int, long, np.integer)):
            raise FlatteningError(n, self)
        if n >= 0x80000000 or n < -0x80000000:
            raise ValueError("out of range for type i: {0}".format(n))
        return struct.pack(endianness + 'i', n), self

registerType(int, LRInt())
registerType(np.int32, LRInt())
registerType(np.int64, LRInt())


class LRWord(LRType, Singleton):
    """An unsigned 32-bit integer."""
    tag = 'w'
    width = 4
    def __unflatten__(self, s, endianness):
        return long(struct.unpack(endianness + 'I', s.get(4))[0])

    def __flatten__(self, n, endianness):
        if not isinstance(n, (int, long, np.integer)):
            raise FlatteningError(n, self)
        if n > 0xFFFFFFFF or n < 0:
            raise ValueError("out of range for type w: {0}".format(n))
        return struct.pack(endianness + 'I', n), self

registerType(long, LRWord())
registerType(np.uint32, LRWord())
registerType(np.uint64, LRWord())


class LRStr(LRType, Singleton):
    """A string of bytes prefixed by a 32-bit length field."""
    tag = 's'
    isFixedWidth = False

    def __width__(self, s, endianness):
        width = struct.unpack(endianness + 'i', s.get(4))[0]
        s.skip(width)
        return 4 + width

    def __unflatten__(self, s, endianness):
        n = struct.unpack(endianness + 'i', s.get(4))[0]
        return s.get(n)

    def __flatten__(self, s, endianness):
        if isinstance(s, unicode):
            s = s.encode('UTF-8')
        if not isinstance(s, str):
            raise FlatteningError(s, self)
        return struct.pack(endianness + 'I', len(s)) + s, self

registerType(str, LRStr())
registerType(unicode, LRStr())

class LRBytes(LRType, Singleton):
    """A raw 8-bit byte string."""
    tag = 'y'
    isFixedWidth = False

    def __width__(self, s, endianness):
        width = struct.unpack(endianness + 'i', s.get(4))[0]
        s.skip(width)
        return 4 + width

    def __unflatten__(self, s, endianness):
        n = struct.unpack(endianness + 'i', s.get(4))[0]
        return s.get(n)

    def __flatten__(self, s, endianness):
        if not isinstance(s, str):
            raise FlatteningError(s, self)
        return struct.pack(endianness + 'I', len(s)) + s, self

# Since bytes is an alias for str in python 2.7, don't enable this
# until we are ready to cut over.
# registerType(bytes, LRBytes)

def timeOffset():
    now = time.time()
    return (datetime.datetime(1904, 1, 1)
            - datetime.datetime.utcfromtimestamp(now)
            + datetime.datetime.fromtimestamp(now))

class LRTime(LRType, Singleton):
    """A timestamp in LabRAD format.

    Timestamp format comes from LabVIEW, and consists of two
    64-bit integers, representing seconds and fractions of a
    second since Jan. 1, 1904, UTC.
    """

    tag = 't'
    width = 16

    def __unflatten__(self, s, endianness):
        secs, us = struct.unpack(endianness + 'QQ', s.get(16))
        us = float(us) / pow(2, 64) * pow(10, 6)
        t = timeOffset() + datetime.timedelta(seconds=secs, microseconds=us)
        return t

    def __flatten__(self, t, endianness):
        diff = t - timeOffset()
        secs = diff.days * (60 * 60 * 24) + diff.seconds
        us = long(float(diff.microseconds) / pow(10, 6) * pow(2, 64))
        return struct.pack(endianness + 'QQ', secs, us), self

registerType(datetime.datetime, LRTime())

'''
class LRFloat(LRType, Singleton):
    tag = 'f'
    width = 8

    def __unflatten__(self, s, endianness):
        return struct.unpack(endianness + 'd', s.get(8)[0])

    def __flatten__(self, f, endianness):
        return struct.pack(endianness + 'd', f)

registerType(float, LRFloat)
'''

class LRValue(LRType):
    """Represents the type of a real number that carries units."""

    tag = 'v'
    width = 8

    # Types which can be flattened by LRValue as 'v[]' via coercion to
    # float. See __flatten__.
    CASTABLE_TYPES = [int, long]

    def __init__(self, unit=None):
        if isinstance(unit, U.Unit):
            unit = str(unit)
        self.unit = unit

    def __str__(self):
        if self.unit is None:
            return self.tag
        else:
            return self.tag + '[%s]' % self.unit

    def __repr__(self):
        return self.__class__.__name__ + '(%r)' % self.unit

    @classmethod
    def __parse__(cls, s):
        return cls(parseUnits(s))

    def __eq__(self, other):
        return type(self) == type(other) and self.unit == other.unit

    def __le__(self, other):
        # this method is a bit funky.  The <= relationship determines
        # which types are allowed to be coerced in flattening.  If the
        # other type is also a value, we allow this if we have a unit,
        # or the other value does not.  In other words, the only case
        # disallowed is the case where we have no unit but the other
        # type does.  This prevents the unit from getting lost in the coercion.
        if type(other) == LRAny:
            return True
        if type(self) != type(other):
            return False
        # If other is 'v', then any variant of 'v' is allowed. For example, if
        # we are 'v[]' or 'v[Hz]', we are allowed to pass to a setting
        # advertising 'v'.
        if other.unit is None:
            return True
        if self.unit is None:
            msg = "Unreachable: no python object should get %s as type"%(self,)
            raise TypeError(msg)
        # We have a unit, and the other guy has a unit, so make sure our units
        # are compatible.
        return U.Unit(self.unit).isCompatible(U.Unit(other.unit))

    def isFullySpecified(self):
        return self.unit is not None

    def __unflatten__(self, s, endianness):
        v = struct.unpack(endianness + 'd', s.get(8))[0]
        if self.unit is not None:
            v = Value(v, self.unit)
        return v

    @classmethod
    def __lrtype__(cls, v):
        """
        Get a LRValue instance for a python object.

        We handle the follwing types:
            float -> LRValue('')
            labrad.units.Value(unit) -> LRValue(str(unit))
        """
        if isinstance(v, U.WithUnit):
            return cls(v.unit)
        elif isinstance(v, (float, U.DimensionlessFloat)):
            return cls('')
        else:
            raise TypeError("No %s type for %s"%(cls, v))

    def __flatten__(self, v, endianness):
        v = self.__check_units__(v)
        v = float(v)
        # If we are an LRValue(None), switch to LRValue('') so that wire
        # data has correct type tag, eg. 'v[]' instead of 'v'.
        if self.unit is None:
            self.unit = ''
        return struct.pack(endianness + 'd', float(v)), self

    def __check_units__(self, v):
        # TODO: implement full labrad unit conversion semantics in pylabrad
        # If v is not a unit-full object, only accept if our unit is
        # None or ''.
        if not isinstance(v, U.WithUnit):
            if self.unit is not None and self.unit != '':
                raise FlatteningError(v, self)
        else:
            # v has units, so convert to required units.
            v = v[self.unit]
        return v

registerTypeFunc((float, Value), LRValue.__lrtype__)


class LRComplex(LRValue):
    """Represents the type of a complex number that carries units."""

    tag = 'c'
    width = 16

    def __unflatten__(self, s, endianness):
        real, imag = struct.unpack(endianness + 'dd', s.get(16))
        c = complex(real, imag)
        if self.unit is not None:
            c = Complex(c, self.unit)
        return c

    def __flatten__(self, c, endianness):
        c = self.__check_units__(c)
        c = complex(c)
        if self.unit is None:
            self.unit = ''
        return struct.pack(endianness + 'dd', c.real, c.imag), self

    @classmethod
    def __lrtype__(cls, c):
        if isinstance(c, U.WithUnit):
            return cls(c.unit)
        elif isinstance(c, (complex, U.DimensionlessComplex)):
            return cls('')
        else:
            raise TypeError("No %s type for %s"%(cls, c))

registerTypeFunc((complex, Complex), LRComplex.__lrtype__)


class LRCluster(LRType):
    """A cluster type for bundling pieces of data together."""

    tag = '('

    def __init__(self, *items):
        self.items = items

    def __str__(self):
        return '(%s)' % ''.join(str(i) for i in self.items)

    def __repr__(self):
        contents = '(%s)' % ', '.join(repr(i) for i in self.items)
        return self.__class__.__name__ + contents

    @classmethod
    def __parse__(cls, s):
        items = []
        while len(s) and s[0] != ')':
            items.append(parseSingleType(s))
        if s.get(1) != ')':
            raise Exception('Unbalanced parentheses in cluster.')
        return cls(*items)

    @classmethod
    def __lrtype__(cls, c):
        return cls(*[getType(i) for i in c])

    def __len__(self):
        return len(self.items)

    def __getitem__(self, key):
        return self.items[key]

    def __le__(self, other):
        """Test whether this type is more specific than another.

        Compatibility requires that both clusters have the same length,
        and all of our items are more specific than the corresponding
        items in the other cluster.
        """
        return (type(other) == LRAny or
                (type(self) == type(other) and
                 len(self.items) == len(other.items) and
                 all(s <= o for s, o in zip(self.items, other.items))))

    def isFullySpecified(self):
        return all(t.isFullySpecified() for t in self.items)

    @property
    def isFixedWidth(self):
        if hasattr(self, '_isFixedWidth'):
            return self._isFixedWidth
        self._isFixedWidth = all(item.isFixedWidth for item in self.items)
        if self._isFixedWidth:
            self.width = sum(item.width for item in self.items)

    def __width__(self, s, endianness):
        return sum(item.__width__(s, endianness) for item in self.items)

    def __unflatten__(self, s, endianness):
        """Unflatten items into a python tuple."""
        return tuple(unflatten(s, t, endianness) for t in self.items)

    def partial_unflatten(self, s, endianness):
        """Unflatten a cluster into a tuple of FlatData objects"""
        s = Buffer(s)
        s2 = Buffer(s)
        item_sizes = [item.__width__(s, endianness) for item in self.items]
        return tuple(FlatData(s2.get(item_size), item, endianness)
                     for (item, item_size) in zip(self.items, item_sizes))

    def __flatten__(self, c, endianness):
        """Flatten python tuple to LabRAD cluster."""
        if len(c) == 0:
            raise FlatteningError('Cannot flatten zero-length clusters')
        if len(c) != len(self.items):
            raise FlatteningError('Cannot flatten %s to %s' % (c, self.items))
        if LRAny() in self.items:
            strs = []
            items = []
            for t, elem in zip(self.items, c):
                if t == LRAny():
                    flat = flatten(elem, endianness=endianness)
                    strs.append(flat.bytes)
                    items.append(flat.tag)
                else:
                    flat = t.flatten(elem, endianness)
                    strs.append(flat.bytes)
                    items.append(flat.tag)
            self.items = items # warning: type mutated here
            return ''.join(strs), self
        return ''.join(t.flatten(elem, endianness).bytes for t, elem in zip(self.items, c)), self

registerTypeFunc(tuple, LRCluster.__lrtype__)


#class LRChoice(LRType):
#    """Represents a choice among different data types."""
#
#    tag = '<'
#
#    def __init__(self, *choices):
#        self.choices = choices
#
#    def __str__(self):
#        return '<%s>' % '|'.join(str(i) for i in self.choices)
#
#    def __repr__(self):
#        contents = '<%s>' % '|'.join(repr(i) for i in self.choices)
#        return self.__class__.__name__ + contents
#
#    @classmethod
#    def __parse__(cls, s):
#        choices = []
#        while len(s) and s[0] != '>':
#            cluster = []
#            while len(s) and s[0] != '|':
#                cluster.append(parseSingleType(s))
#            if len(cluster) == 0:
#                raise Exception('Type choices cannot be empty')
#            elif len(cluster) == 1:
#                choices.append(cluster[0])
#            else:
#                choices.append(LRCluster(*choices))
#            if s.get(1) != '|': # pop off the '|'
#                raise Exception('Expected "|" to delimit type choices.')
#        if s.get(1) != '>':
#            raise Exception('Unbalanced brackets in choice.')
#        return cls(*choices)
#
#    #@classmethod
#    #def __lrtype__(cls, c):
#    #    return cls(*[getType(i) for i in c])
#
#    def __len__(self):
#        return len(self.choices)
#
#    def __getitem__(self, key):
#        return self.choices[key]
#
#    def __le__(self, other):
#        """Test whether this type is more specific than another.
#
#        Compatibility requires that both clusters have the same length,
#        and all of our items are more specific than the corresponding
#        items in the other cluster.
#        """
#        # FIXME: implement this
#        return (type(other) == LRAny or
#                all(s <= other for s in self.choices))
#
#    def isFullySpecified(self):
#        return len(self.choices) == 1
#
#    @property
#    def isFixedWidth(self):
#        return False
#        #if hasattr(self, '_isFixedWidth'):
#        #    return self._isFixedWidth
#        #self._isFixedWidth = all(item.isFixedWidth for item in self.items)
#        #if self._isFixedWidth:
#        #    self.width = sum(item.width for item in self.items)
#
#    def __width__(self, s, endianness):
#        return sum(item.__width__(s, endianness) for item in self.items)
#
#    def __unflatten__(self, s, endianness):
#        """Unflatten items into a python tuple."""
#        return tuple(unflatten(s, t, endianness) for t in self.items)
#
#    def __flatten__(self, c, endianness):
#        """Flatten python tuple to LabRAD cluster."""
#        if len(c) == 0:
#            raise FlatteningError('Cannot flatten zero-length clusters')
#        if len(c) != len(self.items):
#            raise FlatteningError('Cannot flatten %s to %s' % (c, self.items))
#        if LRAny() in self.items:
#            strs = []
#            items = []
#            for t, elem in zip(self.items, c):
#                if t == LRAny():
#                    s, t = flatten(elem, endianness=endianness)
#                    strs.append(s)
#                    items.append(t)
#                else:
#                    flat = t.flatten(elem, endianness)
#                    strs.append(flat.bytes)
#                    items.append(flat.tag)
#            self.items = items # warning: type mutated here
#            return ''.join(strs)
#        return ''.join(t.flatten(elem, endianness) for t, elem in zip(self.items, c))


class LRList(LRType):
    """A multidimensional rectangular array type."""

    tag = '*'

    ARRAY_TYPES = [LRValue(), LRComplex(), LRInt(), LRWord(), LRBool()]

    def __init__(self, elem=LRNone(), depth=1):
        self.elem = elem
        self.depth = depth

    def __str__(self):
        depth = str(self.depth) if self.depth != 1 else ''
        elem = str(self.elem) if self.elem is not None else '_'
        return '*%s%s' % (depth, elem)

    def __repr__(self):
        contents = '(%r, depth=%r)' % (self.elem, self.depth)
        return self.__class__.__name__ + contents

    isFixedWidth = False

    def __width__(self, s, endianness):
        n, elem = self.depth, self.elem
        dims = struct.unpack(endianness + ('i'*n), s.get(4*n))
        size = reduce(lambda x, y: x*y, dims)
        if elem is None:
            width = 0
        elif elem.isFixedWidth:
            width = size * elem.width
        else:
            newBuf = Buffer(s)
            width = sum(elem.__width__(newBuf, endianness) for _ in xrange(size))
        s.skip(width)
        return 4*n + width

    @classmethod
    def __parse__(cls, s):
        s = s.strip(WHITESPACE)
        # get the list dimensionality
        n = parseNumber(s)
        if n == 0:
            raise Exception('Cannot create 0-dimensional list.')
        n = n or 1 # if there were no digits, make a 1D list
        s = s.strip(WHITESPACE)
        if s[:1] == '_':
            t = LRNone() # empty list
            s.get() # drop underscore
        else:
            t = parseSingleType(s)
        return cls(t, n)

    @classmethod
    def __lrtype__(cls, L):
        """
        Return a type object for L.
        """
        depth, temp = 1, L
        while len(temp) and isinstance(temp[0], list):
            depth, temp = depth+1, temp[0]

        def iterND(ls):
            if len(ls) and isinstance(ls[0], list):
                return itertools.chain(*(iterND(i) for i in ls))
            else:
                return iter(ls)

        t = LRAny()
        size = 0
        for elem in iterND(L):
            size += 1
            elem_t = getType(elem)
            if elem_t <= t:
                t = elem_t
            if t.isFullySpecified():
                break
        if size == 0:
            t = LRNone()
        return cls(t, depth)

    @classmethod
    def __lrtype_array__(cls, L):
        #numpy.dtype overloads the == operator, and 'in'
        #uses ==, so our use of in here should be ok
        if L.dtype == 'bool': t = LRBool()
        elif L.dtype in ['>i4', '<i4']: t = LRInt()
        elif L.dtype in ['>i8', '<i8']: t = LRInt()
        elif L.dtype in ['>u4', '<u4']: t = LRWord()
        elif L.dtype in ['>u8', '<u8']: t = LRWord()
        elif L.dtype in ['>f8', '<f8']: t = LRValue('')
        elif L.dtype in ['>c16', '<c16']: t = LRComplex('')
        else:
            raise Exception("Can't flatten array of %s" % L.dtype)
        return cls(t, depth=len(L.shape))

    @classmethod
    def __lrtype_ValueArray__(cls, L):
        if L[L.unit].dtype in ['>c16', '<c16']:
            t = LRComplex(L.unit)
        else:
            t = LRValue(L.unit)
        return cls(t, depth=len(L._value.shape))

    def __le__(self, other):
        """Test whether this type is more specific than another.

        We check the list dimensionality (depth) and the element type.
        """
        return (type(other) == LRAny or
                (type(self) == type(other) and
                 self.depth == other.depth and
                 (other.elem is None or type(other.elem) == LRNone or self.elem <= other.elem)))

    def __eq__(self, other):
        """Test whether this type is equal to another.

        We check the list dimensionality (depth) and the element type.
        """
        return (type(self) == type(other) and
                self.depth == other.depth and
                self.elem == other.elem)

    def isFullySpecified(self):
        return type(self.elem) != LRNone and self.elem.isFullySpecified()

    def __unflatten__(self, s, endianness):
        """Unflatten to numpy array if possible, or else nested python list."""
        n, elem = self.depth, self.elem
        dims = struct.unpack(endianness + ('i'*n), s.get(4*n))
        size = np.prod(dims)

        # Attempt to unflatten as numpy array.
        for t in self.ARRAY_TYPES:
            if elem is not None and elem <= t:
                return self._unflatten_as_array(s, endianness, elem, dims, size)

        # Unflatten as nested python list.
        if self.elem is None or size == 0:
            result = nestedList([], n-1)
        else:
            def unflattenNDlist(s, dims):
                if len(dims) == 1:
                    return [unflatten(s, self.elem, endianness) for _ in xrange(dims[0])]
                else:
                    return [unflattenNDlist(s, dims[1:]) for _ in xrange(dims[0])]
            result = unflattenNDlist(s, dims)
        return LazyList(result)

    def _unflatten_as_array(self, s, endianness, elem, dims, size):
        """Unflatten to numpy array."""
        def make(t, width):
            a = np.fromstring(s.get(size*width), dtype=np.dtype(t))
            if endianness != SYSTEM_BYTE_ORDER:
                a.byteswap(True) # inplace
            return a

        if elem == LRBool(): a = make('bool', 1)
        elif elem == LRInt(): a = make('i4', 4)
        elif elem == LRWord(): a = make('u4', 4)
        elif elem <= LRValue(): a = make('f8', 8)
        elif elem <= LRComplex(): a = make('c16', 16)
        else:
            raise TypeError("Cannot make numpy array with %s"%(elem,))
        a.shape = dims + a.shape[1:] # handle clusters as elements
        if elem <= LRValue() or elem <= LRComplex():
            if elem.unit is not None and elem.unit != '':
                a = U.ValueArray(a, elem.unit)
            else:
                a = U.DimensionlessArray(a)
        return a

    def __flatten__(self, L, endianness):
        """Flatten (nested) python list to LabRAD list.

        Lists must be homogeneous and rectangular.
        """
        if isinstance(L, np.ndarray):
            if L.ndim == 0:
                raise TypeError("can't flatten 0-dimensional array")
            if (self.elem <= LRValue() and
                    not (self.elem.unit is None or self.elem.unit == '')):
                msg = "Can't flatten ndarray to {}".flatten(self)
                raise TypeError(msg)
            return self.__flatten_array__(L, endianness)
        if isinstance(L, U.ValueArray):
            if L.ndim == 0:
                raise TypeError("can't flatten 0-dimensional ValueArray")
            return self.__flatten_ValueArray__(L, endianness)
        if self.elem == LRAny():
            self.elem = self.__lrtype__(L).elem
        lengths = [None] * self.depth
        def flattenNDlist(ls, n=0):
            if lengths[n] is None:
                lengths[n] = len(ls)
            if len(ls) != lengths[n]:
                raise Exception('List is not rectangular.')
            if n+1 == self.depth:
                return ''.join(self.elem.flatten(e, endianness).bytes for e in ls)
            else:
                return ''.join(flattenNDlist(row, n+1) for row in ls)
        flat = flattenNDlist(L)
        lengths = [l or 0 for l in lengths]
        return struct.pack(endianness + ('i' * len(lengths)), *lengths) + flat, self

    def __flatten_array__(self, a, endianness):
        """Flatten numpy array to LabRAD list."""
        shape = a.shape[:self.depth]
        if len(shape) != self.depth:
            raise Exception("Bad array shape.")
        dims = struct.pack(endianness + ('i' * len(shape)), *shape)
        if self.elem == LRAny():
            self.elem = self.__lrtype_array__(a).elem

        # determine what dtype we would like to have
        wanted_dtype = _known_dtypes.get(type(self.elem), None)

        if wanted_dtype is not None:
            if wanted_dtype == 'u4':
                if a.dtype in ['<i8', '>i8']:
                    a = a.astype(endianness + 'u4')
            wanted_dtype = np.dtype(endianness + wanted_dtype)
            if a.dtype != wanted_dtype:
                a_cast = a.astype(wanted_dtype)
                # make sure values don't change in a narrowing cast
                if a.dtype.itemsize > wanted_dtype.itemsize and (not np.all(a_cast == a)):
                    raise Exception("Narrowing typecast loses information while flattening numpy array: dtype={0}, wanted_dtype={1}".format(a.dtype, wanted_dtype))
                a = a_cast
        else:
            elems = (self.elem.flatten(i, endianness=endianness).bytes
                     for i in a.flat)
            return dims + ''.join(elems), self
        return dims + a.tostring(), self

    def __flatten_ValueArray__(self, va, endianness):
        # Convert to appropriate unit and flatten as array
        numericalData = va[self.elem.unit]
        return self.__flatten_array__(numericalData, endianness)


_known_dtypes = {
    LRBool: 'u1',
    LRInt: 'i4',
    LRWord: 'u4',
    LRValue: 'f8',
    LRComplex: 'c16'
}

registerTypeFunc(list, LRList.__lrtype__)
registerTypeFunc(np.ndarray, LRList.__lrtype_array__)
registerTypeFunc(U.ValueArray, LRList.__lrtype_ValueArray__)
registerTypeFunc(U.DimensionlessArray, LRList.__lrtype_array__)


def nestedList(obj, n):
    for _ in range(n):
        obj = [obj]
    return obj


class LRError(LRType):
    """LabRAD error type."""

    tag = 'E'

    def __init__(self, payload=LRNone()):
        self.payload = payload

    def __str__(self):
        if self.payload is LRNone():
            return self.tag
        else:
            return self.tag + str(self.payload)

    def __repr__(self):
        return self.__class__.__name__ + '(%r)' % self.payload

    @classmethod
    def __parse__(cls, s):
        payload = parseSingleType(s)
        return LRError(payload)

    def __eq__(self, other):
        return type(self) == type(other) and self.payload == other.payload

    def __le__(self, other):
        return type(self) == type(other) and self.payload <= other.payload

    def isFullySpecified(self):
        return self.payload.isFullySpecified()

    def __lrtype__(self, E):
        payload = getattr(E, 'payload', None)
        return LRError(getType(payload))

    isFixedWidth = False

    def __width__(self, s, endianness):
        s.skip(4)
        return 4 + LRStr().__width__(s, endianness) + self.payload.__width__(s, endianness)

    def __unflatten__(self, s, endianness):
        """Unflatten to Error type to capture code, message and payload."""
        if self.payload is LRNone():
            t = LRCluster(LRInt(), LRStr())
            code, msg = unflatten(s, t, endianness)
            payload = None
        else:
            t = LRCluster(LRInt(), LRStr(), self.payload)
            code, msg, payload = unflatten(s, t, endianness)
        return Error(msg, code, payload)

    def __flatten__(self, E, endianness):
        """Flatten python Exception to LabRAD error."""
        # TODO add ability to grab tracebacks here, or more information of other types
        code = getattr(E, 'code', 0)
        msg = getattr(E, 'msg', repr(E))
        payload = getattr(E, 'payload', None)
        t = LRCluster(LRInt(), LRStr(), self.payload)
        s, t, _ = flatten((int(code), str(msg), payload), t, endianness)
        self.payload = t.items[2]
        return s, self

registerTypeFunc(Exception, LRError().__lrtype__)


# data types

class Error(Exception):
    """LabRAD base error class.

    Captures the error code and message of a LabRAD error, as well
    as any payload sent along with it.
    """

    # TODO register error classes by code, so remote errors can be reraised

    msg = ''
    payload = None

    def __init__(self, msg=None, code=0, payload=None):
        self.msg = str(msg or self.__doc__)
        self.code = int(code)
        self.payload = payload

    def __str__(self):
        return '(%d) %s [payload=%r]' % (self.code, self.msg, self.payload)

    def __repr__(self):
        return 'Error(code=%r, msg=%r, payload=%r)' % (self.code, self.msg, self.payload)

    def __lrtype__(self):
        return LRError(getType(self.payload))

    def __lrflatten__(self, endianness):
        s, t, _ = flatten((self.code, self.msg, self.payload), endianness=endianness)
        return s, LRError(t.items[2])

class Int(int):
    def __lrtype__(self):
        return LRInt()

class Word(int):
    def __lrtype__(self):
        return LRWord()


class LazyList(list):
    """A proxy object for LabRAD lists.

    List unflattening is not actually done lazily; this class is kept just to
    provide the stub methods below, which now emit deprecation warnings when
    called. This class will be removed in a future release.
    """

    @property
    def aslist(self):
        warnings.warn("LazyList.aslist is deprecated. List unflattening is no "
                      "longer lazy, so this is not needed.")
        return list(self)

    @property
    def astuple(self):
        warnings.warn("LazyList.astuple is deprecated. Use tuple() instead.")
        return tuple(self)

    @property
    def asarray(self):
        warnings.warn("LazyList.asarray is deprecated. Use np.asarray() instead.")
        return np.ndarray(self)


# errors

class FlatteningError(Error):
    """Raised when data cannot be flattened into a valid labrad type."""

    code = 12345

    def __init__(self, data, types=None):
        if types:
            if isinstance(types, list):
                types = [str(t) for t in types]
            else:
                types = str(types)
            self.msg = 'Could not flatten %r to %s.' % (data, types)
        else:
            self.msg = 'Could not flatten %r.' % (data,)



"""LabRAD interface for python

LabRAD is a system for quickly and easily building distributed
instrument control and data analysis applications.  pylabrad
provides a python interface to LabRAD.
"""

classifications = """\
Development Status :: 4 - Beta
Environment :: Console
Environment :: Web Environment
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License (GPL)
Operating System :: OS Independent
Programming Language :: Python
Topic :: Scientific/Engineering"""

import os
from setuptools import setup, find_packages

doclines = __doc__.split('\n')

with open(os.path.join(os.path.dirname(__file__), 'requirements.txt')) as f:
    requirements = f.readlines()

setup(
    name='pylabrad',
    author='Matthew Neeley',
    author_email='mneeley@gmail.com',
    license='http://www.gnu.org/licenses/gpl-2.0.html',
    platforms=['ANY'],

    url='https://github.com/labrad/pylabrad/',
    download_url='',

    description=doclines[0],
    long_description='\n'.join(doclines[2:]),
    classifiers=classifications.split('\n'),

    setup_requires=['setuptools_scm'],
    use_scm_version={
        'write_to': 'labrad/version.py'
    },

    install_requires=requirements,
    provides=['labrad'],
    packages=find_packages(),
    package_data={
        'labrad': ['LICENSE.txt'],
        'labrad.node': ['*.ini'],
    },
    scripts=[],
)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
### BEGIN NODE INFO
[info]
name = Dying Test Server
version = 1.0.0
description = Fails to start; for tests.

[startup]
cmdline = %PYTHON% %FILE%
timeout = 20

[shutdown]
message = 987654321
timeout = 5
### END NODE INFO
"""

from labrad import util
from labrad.server import LabradServer, setting, Signal
from twisted.internet import defer, reactor
from twisted.internet.defer import returnValue
from twisted.python import log

from datetime import datetime

class TestServer(LabradServer):
    """Server to test labrad from python.

    This server provides a number of settings that
    are useful for testing labrad code, network profiling,
    and testing the python labrad library.
    """
    name = 'Dying Test Server'
    testMode = True

    def initServer(self):
        raise Exception()

    def serverConnected(self, data):
        print 'connected:', data

    def serverDisconnected(self, data):
        print 'disconnected:', data

    log = Signal(555, 'log', '(ts)')

    @setting(2)
    def delayed_echo(self, c, data):
        """Echo a packet after a specified delay."""
        yield util.wakeupCall(c['delay'])
        returnValue(data)

    @setting(3)
    def delayed_echo_deferred(self, c, data):
        """Echo a packet after a specified delay."""
        d = defer.Deferred()
        reactor.callLater(c['delay'], d.callback, data)
        return d

    @setting(4, delay=['v[s]', ''], returns=['v[s]'])
    def echo_delay(self, c, delay):
        """Get or set the echo delay."""
        log.msg('Echo delay: %s' % delay)
        self.log((datetime.now(), 'Echo delay: %s' % delay))
        if delay is not None:
            c['delay'] = float(delay)
        return c['delay']

    @setting(40, speed=['v[m/s]', ''], returns=['v[m/s]'])
    def speed(self, c, speed):
        """Get or set the speed."""
        log.msg('Speed: %s' % speed)
        self.log((datetime.now(), 'Speed: %s' % speed))
        if speed is not None:
            c['speed'] = speed
        return c['speed']

    @setting(41)
    def verbose_echo(self, c, data):
        print type(data)
        print repr(data)
        return data

    @setting(5)
    def exc_in_handler(self, c, data):
        log.msg('Exception in handler.')
        raise Exception('Raised in handler.')

    @setting(6)
    def exc_in_subfunction(self, c, data):
        log.msg('Exception in subfunction.')
        owie()

    @setting(7)
    def exc_in_deferred(self, c, data):
        log.msg('Exception in deferred.')
        d = defer.Deferred()
        d.addCallback(owie)
        reactor.callLater(1, d.callback, None)
        return d

    @setting(8)
    def exc_in_errback(self, c, data):
        log.msg('Exception from an errback.')
        d = defer.Deferred()
        reactor.callLater(1, d.errback, Exception('Raised by errback.'))
        return d

    @setting(9)
    def exc_in_inlinecallback(self, c, data):
        log.msg('Exception from an inlineCallback.')
        yield util.wakeupCall(c['delay'])
        raise Exception('Raised in inlineCallback.')

    @setting(10, returns=['s'])
    def bad_return_type(self, c, data):
        return 5

def owie(dummy=None):
    raise Exception('Raised in subfunction.')

__server__ = TestServer()

if __name__ == '__main__':
    from labrad import util
    util.runServer(__server__)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
### BEGIN NODE INFO
[info]
name = Python Test Server
version = 1.0.0
description = Basic python server.

[startup]
cmdline = %PYTHON% %FILE%
timeout = 20

[shutdown]
message = 987654321
timeout = 5
### END NODE INFO
"""

from labrad import types as T, util
from labrad.server import LabradServer, setting
from labrad.units import m, s
from labrad.util import hydrant

from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, returnValue


class TestServer(LabradServer):
    """Server to test labrad from python.

    This server provides a number of settings that
    are useful for testing labrad code, network profiling,
    and testing the python labrad library.
    """
    name = 'Python Test Server'
    testMode = True
    shutdownMessage = 987654321

    @inlineCallbacks
    def initServer(self):
        yield None
        #from registry_wrapper_async import RegistryWrapperAsync
        #self.regWrapper = yield RegistryWrapperAsync.create(self.client, ['', 'Servers', 'Python Test Server'])

    @inlineCallbacks
    def stopServer(self):
        print 'before yield'
        yield None
        #print (yield self.client.manager.convert_units(T.Value(5, 'GHz'), 'Hz'))
        print 'after yield'

    def serverConnected(self, ID, name):
        print 'server connected:', ID, name
        self.checkServerWrappers(name)

    def serverDisconnected(self, ID, name):
        print 'server disconnected:', ID, name
        self.checkServerWrappers(name)

    @inlineCallbacks
    def checkServerWrappers(self, name):
        """Check that server wrappers are up to date with the manager."""
        mgrServers = yield self.client.manager.servers()
        mgrServers = set(s[1] for s in mgrServers)
        cxnServers = set(self.client.servers.keys())
        if cxnServers == mgrServers:
            print 'self.client updated for server', name
        else:
            print 'self.client not properly refreshed:'
            print '  servers that should be disconnected:', list(cxnServers - mgrServers)
            print '  servers that have not been connected:', list(mgrServers - cxnServers)

    def initContext(self, c):
        c['delay'] = 1*s
        c['dict'] = {}

    @setting(2, "Delayed Echo", data='?')
    def delayed_echo(self, c, data):
        """Echo a packet after a specified delay."""
        yield util.wakeupCall(c['delay'][s])
        returnValue(data)

    @setting(3, "Delayed Echo Deferred", data='?')
    def delayed_echo_deferred(self, c, data):
        """Echo a packet after a specified delay."""
        d = defer.Deferred()
        reactor.callLater(c['delay'][s], d.callback, data)
        return d

    @setting(4, "Echo Delay", delay='v[s]', returns='v[s]')
    def echo_delay(self, c, delay=None):
        """Get or set the echo delay."""
        self.log('Echo delay: %s' % delay)
        if delay is not None:
            c['delay'] = delay
        return c['delay']

    @setting(42, "Delayed Echo Wrapper", data='?', returns='?')
    def delayed_echo_wrapper(self, c, data):
        """Echo a packet after a delay just like delayed_echo.

        This tests calling a coroutine from another coroutine.
        """
        rv = yield self.delayed_echo(c, data)
        returnValue(rv)

    @setting(40, "Speed", speed='v[m/s]', returns='v[m/s]')
    def speed(self, c, speed=None):
        """Get or set the speed."""
        self.log('Speed: %s' % speed)
        if speed is not None:
            c['speed'] = speed
        return c['speed']

    @setting(41, "Verbose Echo", data='?')
    def verbose_echo(self, c, data):
        print type(data)
        print repr(data)
        return data

    @setting(5, "Exc in Handler", data='?')
    def exc_in_handler(self, c, data):
        """Raises an exception directly in the handler."""
        self.log('Exception in handler.')
        raise Exception('Raised in handler.')

    @setting(6, "Exc in Subfunction", data='?')
    def exc_in_subfunction(self, c, data):
        """Raises an exception in a subfunction."""
        self.log('Exception in subfunction.')
        owie()

    @setting(7, "Exc in Deferred", data='?')
    def exc_in_deferred(self, c, data):
        """Returns a deferred that fires an exception."""
        self.log('Exception in deferred.')
        d = defer.Deferred()
        d.addCallback(owie)
        reactor.callLater(c['delay'][s], d.callback, None)
        return d

    @setting(8, "Exc in Errback", data='?')
    def exc_in_errback(self, c, data):
        """Returns a deferred whose errback will be called."""
        self.log('Exception from an errback.')
        d = defer.Deferred()
        reactor.callLater(1, d.errback, Exception('Raised by errback.'))
        return d

    @setting(9, "Exc in inlineCallback", data='?')
    def exc_in_inlinecallback(self, c, data):
        """Raises an exception in an inlineCallback."""
        self.log('Exception from an inlineCallback.')
        yield util.wakeupCall(c['delay'][s])
        raise Exception('Raised in inlineCallback.')

    @setting(10, "Bad Return Type", data='?', returns='s')
    def bad_return_type(self, c, data):
        """Returns a value that does not match the declared return type."""
        return 5

    @setting(11, "Get Random Data", tag='s')
    def get_random_data(self, c, tag=None):
        """Get a random bit of data conforming to the specified type tag."""
        if tag is None:
            t = hydrant.randType()
        else:
            t = T.parseTypeTag(tag)
        return hydrant.randValue(t)

    @setting(12, "Get Random Tag")
    def get_random_tag(self, c):
        """Get a random LabRAD type tag."""
        return str(hydrant.randType())

    @setting(100, "Set", unflatten=False, key='s', value='?', returns='')
    def set(self, c, key, value):
        c['dict'][key.unflatten()] = value

    @setting(101, "Get", key='s', returns='?')
    def get(self, c, key):
        print "getting tag: %s, value: %s" % (key, c['dict'][key])
        return c['dict'][key]

    @setting(102, "Keys", returns='*s')
    def keys(self, c):
        return sorted(c['dict'].keys())

    @setting(103, "Set Reversed", value='?', key='s', unflatten=False)
    def set_reversed(self, c, value, key):
        """Same as self.set() with argument order reversed."""
        self.set(c, key, value)

    @setting(1000, path=['', 's'], file=['s'])
    def first_arg_can_have_default_value(self, c, path=None, file=None):
        """Exercise setting decorator fix from issue #242."""
        pass

def owie(dummy=None):
    raise Exception('Raised in subfunction.')

__server__ = TestServer()

if __name__ == '__main__':
    util.runServer(__server__)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.grammar

Grammars for parsing various LabRAD stuff with pyparsing.
"""

from labrad.pyparsing import (Word, Literal, Group,
                              Forward, Optional, alphas, nums, alphanums, stringEnd)

toInt = lambda s, l, t: [int(t[0])]

number = Word(nums).setParseAction(toInt)
name = Word(alphas + '%"\'\xE6\xF8', alphanums + '%"\'\xE6\xF8') # degree and mu

power = Literal('^').suppress()
times = Literal('*').suppress()
divide = Literal('/').suppress()
minus = Literal('-').setResultsName('neg')
one = Literal('1').suppress()
lparen = Literal('(').suppress()
rparen = Literal(')').suppress()
exponent = Optional(minus) + number.setResultsName('num') ^ \
           Optional(minus) + number.setResultsName('num') + divide \
                           + number.setResultsName('denom')

single_unit = name.setResultsName('name') + Optional(power + exponent)
bare_unit = Group(single_unit).setResultsName('posexp', listAllMatches=True)
num_unit = Group(times + single_unit).setResultsName('posexp', listAllMatches=True)
denom_unit = Group(divide + single_unit).setResultsName('negexp', listAllMatches=True)

later_units = Forward()
later_units << (num_unit + Optional(later_units) |
                denom_unit + Optional(later_units))

unit = Forward()
unit << ((bare_unit | one + denom_unit) + Optional(later_units) + stringEnd)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.constants

Important constants that show up throughout the code.
"""

import os
import labrad.crypto

DEFAULT_TLS = 'starttls' if labrad.crypto.TLS else 'off'

def check_tls_mode(tls_mode):
    """Check that provided tls mode is valid and convert to canonical form.

    Args:
        tls_mode (string): One of 'on', 'off', 'starttls', or 'starttls-force'.
            If 'on', we use TLS for the initial session establishment.
            If 'off', no TLS is used (for connecting to a legacy manager that
            does not support TLS). If 'starttls', the connection is initially
            unencrypted and will then be upgraded to a TLS connection after the
            initial handshake when connecting to a remote manager.
            If 'starttls-force', the client will behave like starttls mode but
            the connection will be upgraded to TLS even when connecting to the
            manager on localhost.

    Returns (string): tls mode in canonical (lowercase) form.

    Raises:
        ValueError: an invalid tls mode was specified
    """
    tls_mode = tls_mode.lower()
    if tls_mode not in ['on', 'off', 'starttls', 'starttls-force']:
        raise ValueError("tls mode must be one of 'on', 'off', 'starttls', or "
                         "'starttls-force'. got: '{}'".format(tls_mode))
    return tls_mode

# defaults for the labrad manager
MANAGER_ID = 1
MANAGER_PORT = int(os.environ.get('LABRADPORT', 7682))
MANAGER_TLS = check_tls_mode(os.environ.get('LABRAD_TLS', DEFAULT_TLS))
MANAGER_PORT_TLS = int(os.environ.get('LABRAD_TLS_PORT', 7643))
MANAGER_HOST = os.environ.get('LABRADHOST', 'localhost')
USERNAME = os.environ.get('LABRADUSER', None)
PASSWORD = os.environ.get('LABRADPASSWORD', None)

# settings IDs on the LabRAD manager
SERVERS_LIST = 1
SETTINGS_LIST = 2
LOOKUP = 3
HELP = 10
MESSAGE_SUBSCRIBE = 60

# default timeout in seconds
TIMEOUT = 30

# node controller
HTTP_PORT = int(os.environ.get('LABRADHTTPPORT', 7667))



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
labrad.gpib

Superclass of GPIB device servers.
"""

from labrad import types as T, constants as C, util, errors
from labrad.devices import DeviceWrapper, DeviceServer, DeviceLockedError
from labrad.server import LabradServer, setting
from labrad.support import MultiDict

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks, returnValue

class GPIBDeviceWrapper(DeviceWrapper):
    """A wrapper for a gpib device."""

    @inlineCallbacks
    def connect(self, gpib, addr):
        """Connect to this device.

        We set the address and timeout in the context reserved
        for talking to this device.  Then call initialize, which may
        be overridden in subclasses.
        """
        self.gpib = gpib # wrapper for the gpib server
        self.addr = addr
        self._context = gpib.context() # create a new context for this device
        self._timeout = T.Value(C.TIMEOUT, 's')

        # set the address and timeout in this context
        p = self._packet()
        p.address(self.addr)
        p.timeout(self._timeout)
        yield p.send()

        # do device-specific initialization
        yield self.initialize()

    def _packet(self):
        return self.gpib.packet(context=self._context)

    @inlineCallbacks
    def timeout(self, seconds):
        """Set the GPIB timeout for this device."""
        self._timeout = T.Value(seconds, 's')
        p = self._packet()
        p.timeout(self._timeout)
        yield p.send()

    @inlineCallbacks
    def query(self, query, bytes=None, timeout=None):
        """Query this GPIB device."""
        p = self._packet()
        if timeout is not None:
            p.timeout(timeout)
        p.query(query)
        if timeout is not None:
            p.timeout(self._timeout)
        resp = yield p.send()
        returnValue(resp.query)

    @inlineCallbacks
    def write(self, s, timeout=None):
        """Write a string to the device."""
        p = self._packet()
        if timeout is not None:
            p.timeout(timeout)
        p.write(s)
        if timeout is not None:
            p.timeout(self._timeout)
        resp = yield p.send()
        returnValue(resp.write)

    @inlineCallbacks
    def write_raw(self, s, timeout=None):
        """Write a string to the device."""
        p = self._packet()
        if timeout is not None:
            p.timeout(timeout)
        p.write_raw(s)
        if timeout is not None:
            p.timeout(self._timeout)
        resp = yield p.send()
        returnValue(resp.write_raw)


    @inlineCallbacks
    def read(self, bytes=None, timeout=None):
        """Read a string from the device."""
        p = self._packet()
        if timeout is not None:
            p.timeout(timeout)
        p.read(bytes)
        if timeout is not None:
            p.timeout(self._timeout)
        resp = yield p.send()
        returnValue(resp.read)

    @inlineCallbacks
    def read_raw(self, bytes=None, timeout=None):
        """Read a string from the device."""
        p = self._packet()
        if timeout is not None:
            p.timeout(timeout)
        p.read_raw(bytes)
        if timeout is not None:
            p.timeout(self._timeout)
        resp = yield p.send()
        returnValue(resp.read_raw)

    def initialize(self):
        """Called when we first connect to the device.

        Override this in subclasses to perform device-specific
        initialization and to synchronize the wrapper state with
        the device state.
        """

class GPIBDeviceServer(DeviceServer):
    """A server for a GPIB device.

    Creates a GPIBDeviceWrapper for each device it finds that
    is appropriately named.  Provides standard settings for listing
    devices, selecting a device for the current context, and
    refreshing the list of devices.  Also, allows us to read from,
    write to, and query the selected GPIB device directly.

    2013 October 22 - Daniel Sank
    I'm pretty sure this is obsolte and you should use the GPIBManagedServer
    """
    name = 'Generic GPIB Device Server'
    deviceName = 'Generic GPIB Device'
    deviceWrapper = GPIBDeviceWrapper

    def serverConnected(self, ID, name):
        """Refresh devices when a gpib server comes on line."""
        if 'gpib' in name.lower():
            self.refreshDeviceList()

    def serverDisconnected(self, ID, name):
        """Refresh devices when a gpib server goes off line."""
        if 'gpib' in name.lower():
            self.refreshDeviceList()

    @inlineCallbacks
    def findDevices(self):
        """Find all available matching GPIB devices."""
        searches = [self._findDevicesForServer(srv)
                    for srv in _gpibServers(self.client)]
        found = []
        for search in searches:
            found += yield search
        returnValue(found)

    @inlineCallbacks
    def _findDevicesForServer(self, srv):
        """Find matching devices on a given server."""
        devices = yield srv.list_devices()
        found = []
        for address, deviceName in zip(devices, ['<UNKNOWN>']*len(devices)):
            if deviceName == self.deviceName:
                name = _getDeviceName(srv, address)
                found.append((name, (srv, address), {}))
        returnValue(found)

    # server settings

    @setting(1001, 'GPIB Write', string='s', returns='*b')
    def gpib_write(self, c, string):
        """Write a string to the device over GPIB."""
        return self.selectedDevice(c).write(string)

    @setting(1002, 'GPIB Read', bytes='w', returns='s')
    def gpib_read(self, c, bytes=None):
        """Read a string from the device over GPIB."""
        return self.selectedDevice(c).read(bytes)

    @setting(1003, 'GPIB Query', query='s', returns='s')
    def gpib_query(self, c, query):
        """Write a string over GPIB and read the response."""
        return self.selectedDevice(c).query(query)

class ManagedDeviceServer(LabradServer):
    """A server for devices.

    Creates a DeviceWrapper for each device it finds, based on a
    user-provided function.  Provides standard settings for listing
    devices, selecting a device for the current context, and
    refreshing the list of devices.  Also, provides for device-locking
    with timeouts.

    Device names and associated wrappers can be specified in two ways
    1 (old way)
    Give a device name or list of device names:
    deviceName = "nameOfDevice" eg. "Acme XL123" or
    deviceName = ["nameOfDevice", "nameOfOtherDevice",...]
    and also give a single device wrapper.
    deviceWrapper=<a device Wrapper (sub)class>
    With this method the same device wrapper is used for all detected
    devices, regardless of eg. model. This works if all models use
    the same SCPI commands.
    2 (new better way)
    Give a dict mapping device names to wrappers
    deviceWrappers = {"deviceName":wrapperForThisDevice,...}
    This allows you to use the same server for eg. devices of the same
    general type but from different manufacturers or of different
    models.

    1. Old way example
    deviceName = "Acme Widget"
    deviceWrapper = AcmeWidgetWrapper
    2. New way example
    deviceWrappers={"Acme Widget": AcmeWidgetExample}

    Optionally specify a device specific identication function
    deviceIdentFunc = 'identify_device'
    """
    name = 'Generic Device Server'
    deviceManager = 'Device Manager'

    # Default device name and wrapper for backwards compatibility with servers
    # written before we supported multiple different device types, and which
    # do not explicitly set deviceWrapper and/or deviceName.
    deviceName = 'Generic Device'
    deviceWrapper = DeviceWrapper

    messageID = 21436587

    def __init__(self):
        #Backward compatibility for servers that don't use a
        #deviceWrappers dict
        if not hasattr(self, 'deviceWrappers'):
            names = self.deviceName
            if isinstance(names, str):
                names = [names]
            self.deviceWrappers = dict((name, self.deviceWrapper) for name in names)
        LabradServer.__init__(self)

    @inlineCallbacks
    def initServer(self):
        self.devices = MultiDict() # aliases -> device
        self.device_guids = {} # name -> guid
        self._next_guid = 0
        # register a message handler for connect/disconnect messages
        handler = lambda c, data: self.handleDeviceMessage(*data)
        self._cxn.addListener(handler, ID=self.messageID)
        if self.deviceManager in self.client.servers:
            yield self.connectToDeviceManager()

    @inlineCallbacks
    def stopServer(self):
        if hasattr(self, 'devices'):
            ds = [defer.maybeDeferred(dev.shutdown)
                  for dev in self.devices.values()]
            yield defer.DeferredList(ds, fireOnOneErrback=True)

    def makeDeviceName(self, device, server, address):
        return server + ' - ' + address

    @inlineCallbacks
    def handleDeviceMessage(self, device, server, address, isConnected=True):
        print 'Device message:', device, server, address, isConnected
        name = self.makeDeviceName(device, server, address)
        if isConnected: # add device
            if name in self.devices:
                return # we already have this device
            if name in self.device_guids:
                # we've seen this device before
                # so we'll reuse the old guid
                guid = self.device_guids[name]
            else:
                guid = self.device_guids[name] = self._next_guid
                self._next_guid += 1
            dev = self.deviceWrappers[device](guid, name)
            yield self.client.refresh()
            yield dev.connect(self.client[server], address)
            self.devices[guid, name] = dev

        else: # remove device
            yield self.removeDevice(name)

    @inlineCallbacks
    def removeDevice(self, name):
        # we delete the device, but not its guid, so that
        # if this device comes back, users who have
        # selected it by guid will reconnect seamlessly
        if name not in self.devices:
            return
        dev = self.devices[name]
        del self.devices[name]
        try:
            yield dev.shutdown()
        except Exception, e:
            self.log('Error while shutting down device "%s": %s' % (name, e))

    @inlineCallbacks
    def connectToDeviceManager(self):
        """

        """
        yield self.client.refresh()
        manager = self.client[self.deviceManager]
        #If we have a device identification function register it with the device manager
        if hasattr(self, 'deviceIdentFunc'):
            yield manager.register_ident_function(self.deviceIdentFunc)
        #Register ourself as a server who cares about devices
        devs = yield manager.register_server(self.deviceWrappers.keys(), self.messageID)
        # the devs list is authoritative any devices we have
        # that are _not_ on this list should be removed
        names = [self.makeDeviceName(*dev[:3]) for dev in devs]
        additions = [self.handleDeviceMessage(*dev) for dev in devs]
        deletions = [self.removeDevice(name)
                     for name in self.device_guids.keys()
                     if name not in names]
        yield defer.DeferredList(additions + deletions)

    def serverConnected(self, ID, name):
        if name == self.deviceManager:
            self.connectToDeviceManager()

    def expireContext(self, c):
        """Release selected/locked device when context expires."""
        if 'device' in c:
            alias = c['device']
            try:
                dev = self.devices[alias]
                if dev.lockedInContext(c):
                    dev.unlock(c)
                dev.deselect(c)
            except KeyError:
                pass

    def deviceLists(self):
        """Get parallel lists of device names and IDs."""
        guids = sorted(self.devices.keys())
        names = [self.devices[g].name for g in guids]
        return guids, names

    def selectedDevice(self, context):
        """Get the selected device from the given context."""
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        try:
            key = context['device']
        except KeyError:
            raise errors.DeviceNotSelectedError()
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError()
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()
        return dev

    def selectDevice(self, context, key=None):
        """Select a device in our current context."""
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        if key is None:
            # use the first device
            key = sorted(self.devices.keys())[0]
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError()
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()

        if 'device' in context:
            if context['device'] != dev.guid:
                try:
                    oldDev = self.devices[context['device']]
                except KeyError:
                    pass
                else:
                    # we're trying to select a new device.
                    # make sure to unlock previously selected device
                    if oldDev.lockedInContext(context.ID):
                        oldDev.unlock(context.ID)
                    oldDev.deselect(context)
                context['device'] = dev.guid
                dev.select(context)
        else:
            context['device'] = dev.guid
            dev.select(context)
        return dev

    def deselectDevice(self, context):
        if 'device' in context:
            try:
                oldDev = self.devices[context['device']]
            except KeyError:
                pass
            else:
                # unlock and deselect device
                if oldDev.lockedInContext(context.ID):
                    oldDev.unlock(context.ID)
                oldDev.deselect(context)
            del context['device']

    def getDevice(self, context, key=None):
        if not len(self.devices):
            raise errors.NoDevicesAvailableError()
        if key is None:
            # use the first device
            key = sorted(self.devices.keys())[0]
        try:
            dev = self.devices[key]
        except KeyError:
            raise errors.NoSuchDeviceError()
        if not dev.accessibleFrom(context.ID):
            raise DeviceLockedError()
        return dev

    # server settings

    @setting(1, 'List Devices', returns='*(ws)')
    def list_devices(self, c):
        """Get a list of available devices.

        The list entries have a numerical ID and a string name.
        It is recommended to use the names wherever possible to
        identify devices, since this contains human-readable
        information about the device, however for speed numerical
        IDs may sometimes be desirable.  The ID number for a
        device is persistent, so it can be used even if the device
        disconnects and later reconnects under the same name.
        """
        IDs, names = self.deviceLists()
        return zip(IDs, names)

    @setting(2, 'Select Device',
                key=[': Select first device',
                     's: Select device by name',
                     'w: Select device by ID'],
                returns='s: Name of the selected device')
    def select_device(self, c, key=0):
        """Select a device for the current context."""
        dev = self.selectDevice(c, key=key)
        return dev.name

    @setting(3, 'Deselect Device', returns='')
    def deselect_device(self, c):
        """Deselect a device in the current context."""
        dev = self.deselectDevice(c)

    @setting(1000001, 'Lock Device',
                      timeout=[': Lock the selected device for default time',
                            'v[s]: Lock for specified time'],
                      returns=[''])
    def lock_device(self, c, timeout):
        """Lock a device to be accessible only in this context."""
        dev = self.selectedDevice(c)
        if timeout is not None:
            timeout = timeout['s']
        dev.lock(c.ID, timeout)

    @setting(1000002, 'Release Device', returns='')
    def release_device(self, c):
        """Release the lock on the currently-locked device."""
        dev = self.selectedDevice(c)
        dev.unlock(c.ID)

class GPIBManagedServer(ManagedDeviceServer):
    """A server for a GPIB device.

    Creates a GPIBDeviceWrapper for each device it finds that
    is appropriately named.  Provides standard settings for listing
    devices, selecting a device for the current context, and
    refreshing the list of devices.  Also, allows us to read from,
    write to, and query the selected GPIB device directly.
    """
    name = 'Generic GPIB Device Server'
    deviceManager = 'GPIB Device Manager'
    deviceName = 'Generic GPIB Device'
    deviceWrapper = GPIBDeviceWrapper
    # server settings

    @setting(1001, 'GPIB Write', string='s', timeout='v[s]', returns='')
    def gpib_write(self, c, string, timeout=None):
        """Write a string to the device over GPIB."""
        return self.selectedDevice(c).write(string, timeout)

    @setting(1002, 'GPIB Read', bytes='w', timeout='v[s]', returns='s')
    def gpib_read(self, c, bytes=None, timeout=None):
        """Read a string from the device over GPIB."""
        return self.selectedDevice(c).read(bytes, timeout)

    @setting(1003, 'GPIB Query', query='s', timeout='v[s]', returns='s')
    def gpib_query(self, c, query, timeout=None):
        """Write a string over GPIB and read the response."""
        return self.selectedDevice(c).query(query, timeout=timeout)

def _gpibServers(cxn):
    """Get a list of available GPIB servers."""
    gpibs = []
    for name in cxn.servers:
        srv = cxn.servers[name]
        if 'gpib_bus' in name and 'list_devices' in srv.settings:
            gpibs.append(srv)
    return gpibs

def _getDeviceName(server, deviceID):
    """Create a name for a device on a particular server."""
    return '%s - %s' % (server.name, deviceID)



# Copyright (C) 2007  Matthew Neeley
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from twisted.internet.defer import inlineCallbacks, returnValue

from labrad import constants as C, types as T

class AsyncManager:
    """Provide an asynchronous interface to basic manager settings."""

    ID = C.MANAGER_ID

    def __init__(self, cxn):
        self.cxn = cxn

    def _send(self, packet, *args, **kw):
        """Send a request to the manager."""
        return self.cxn.sendRequest(self.ID, packet, *args, **kw)

    @inlineCallbacks
    def _getIDList(self, setting, data=None):
        resp = yield self._send([(setting, data)])
        names = self._reorderIDList(resp[0][1])
        returnValue(names)

    def _reorderIDList(self, L):
        return [(name, ID) for ID, name in L]

    def getServersList(self):
        """Get a list of connected servers."""
        return self._getIDList(T.flatten(C.SERVERS_LIST, 'w'))

    @inlineCallbacks
    def getServerInfo(self, serverID):
        """Get information about a server."""
        packet = [(C.HELP, T.flatten(serverID, 'w')),
                  (C.SETTINGS_LIST, T.flatten(serverID, 'w'))]
        resp = yield self._send(packet)
        descr, notes = resp[0][1]
        settings = self._reorderIDList(resp[1][1])
        returnValue((descr, notes, settings))

    @inlineCallbacks
    def getServerInfoWithSettings(self, serverID):
        """Get information about a server, including all of its settings."""
        packet = [(C.HELP, T.flatten(serverID, 'w')),
                  (C.SETTINGS_LIST, T.flatten(serverID, 'w'))]
        resp = yield self._send(packet)
        descr, notes = resp[0][1]
        settings = resp[1][1]
        packet = [(C.HELP, T.flatten((serverID, ID), 'ww')) for ID, name in settings]
        resp = yield self._send(packet)
        settingList = []
        for s, r in zip(settings, resp):
            ID, name = s
            descr, accepts, returns, notes = r[1]
            settingList.append((name, ID, (descr, accepts, returns, notes)))
        returnValue((descr, notes, settingList))

    def getSettingsList(self, serverID):
        """Get a list of settings for a server."""
        return self._getIDList(C.SETTINGS_LIST, T.flatten(serverID, 'w'))

    @inlineCallbacks
    def getSettingInfo(self, serverID, settingID):
        """Get information about a setting."""
        packet = [(C.HELP, T.flatten((serverID, settingID), 'ww'))]
        resp = yield self._send(packet)
        description, accepts, returns, notes = resp[0][1]
        returnValue((description, accepts, returns, notes))

    @inlineCallbacks
    def getSettingInfoByName(self, serverID, settingName):
        """Get information about a setting using its name."""
        packet = [(C.HELP, T.flatten((serverID, settingName), 'ws')),
                  (C.LOOKUP, T.flatten((serverID, settingName), 'ws'))]
        resp = yield self._send(packet)
        description, accepts, returns, notes = resp[0][1]
        ID = resp[1][1][1]
        returnValue((description, accepts, returns, notes, ID))

    @inlineCallbacks
    def subscribeToNamedMessage(self, name, ID, enable=True):
        """Subscribe to or stop a named message."""
        packet = [(C.MESSAGE_SUBSCRIBE, T.flatten((name, ID, enable), 'swb'))]
        returnValue((yield self._send(packet)))



import Queue
import random

import pytest

import labrad
from labrad.server import LabradServer, setting
from labrad import util


def test_server_expire_context_method_is_called():
    """Ensure that server's expireContext method is called when client disconnects."""
    queue = Queue.Queue()

    class TestServer(LabradServer):
        name = "TestServer"

        def expireContext(self, c):
            queue.put(c.ID)

        @setting(100)
        def echo(self, c, data):
            return data

    with util.syncRunServer(TestServer()):
        with labrad.connect() as cxn:
            # create a random context owned by this connection
            request_context = (cxn.ID, random.randint(0, 0xFFFFFFFFL))
            cxn.testserver.echo('hello, world!', context=request_context)
        expired_context = queue.get(block=True, timeout=1)
        assert expired_context == request_context


def test_server_init_failure_is_propagated():

    class InitError(Exception):
        pass

    class DyingServer(LabradServer):
        name = "Dying Server"
        def initServer(self):
            raise InitError()

    with pytest.raises(InitError):
        with util.syncRunServer(DyingServer()):
            pass


if __name__ == '__main__':
    pytest.main(['-v', __file__])
